from itertools import groupby
import logging
import re

from cinderclient import exceptions as cinder_exceptions
from django.db import transaction
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.decorators import method_decorator
from keystoneclient import exceptions as keystone_exceptions
from neutronclient.client import exceptions as neutron_exceptions
from novaclient import exceptions as nova_exceptions

from waldur_core.core.utils import create_batch_fetcher
from waldur_core.structure import log_backend_action, SupportedServices
from waldur_core.structure.utils import (
    update_pulled_fields, handle_resource_not_found, handle_resource_update_success)
from waldur_openstack.openstack_base.backend import OpenStackBackendError, BaseOpenStackBackend, reraise

from . import models

logger = logging.getLogger(__name__)


class OpenStackBackend(BaseOpenStackBackend):
    DEFAULTS = {
        'tenant_name': 'admin',
    }

    def check_admin_tenant(self):
        try:
            self.keystone_admin_client
        except keystone_exceptions.AuthorizationFailure:
            return False
        except keystone_exceptions.ClientException as e:
            reraise(e)
        else:
            return True

    def pull_service_properties(self):
        self.pull_flavors()
        self.pull_images()
        self.pull_service_settings_quotas()

    def pull_resources(self):
        self.pull_tenants()

    def pull_subresources(self):
        self.pull_security_groups()
        self.pull_floating_ips()
        self.pull_networks()
        self.pull_subnets()

    def pull_tenants(self):
        keystone = self.keystone_admin_client

        try:
            backend_tenants = keystone.projects.list(domain=self._get_domain())
        except keystone_exceptions.ClientException as e:
            reraise(e)

        backend_tenants_mapping = {tenant.id: tenant for tenant in backend_tenants}

        tenants = models.Tenant.objects.filter(
            state__in=[models.Tenant.States.OK, models.Tenant.States.ERRED],
            service_project_link__service__settings=self.settings,
        )
        for tenant in tenants:
            backend_tenant = backend_tenants_mapping.get(tenant.backend_id)
            if backend_tenant is None:
                handle_resource_not_found(tenant)
                continue

            imported_backend_tenant = models.Tenant(
                name=backend_tenant.name,
                description=backend_tenant.description,
                backend_id=backend_tenant.id,
                state=models.Tenant.States.OK,
            )
            update_pulled_fields(tenant, imported_backend_tenant, models.Tenant.get_backend_fields())
            handle_resource_update_success(tenant)

    def _get_domain(self):
        """ Get current domain """
        keystone = self.keystone_admin_client
        return keystone.domains.find(name=self.settings.domain or 'Default')

    def remove_ssh_key_from_tenant(self, tenant, key_name, fingerprint):
        nova = self.nova_client

        # There could be leftovers of key duplicates: remove them all
        keys = nova.keypairs.findall(fingerprint=fingerprint)
        for key in keys:
            # Remove only keys created with Waldur
            if key.name == key_name:
                nova.keypairs.delete(key)

        logger.info('Deleted ssh public key %s from backend', key_name)

    def _are_rules_equal(self, backend_rule, nc_rule):
        if backend_rule['port_range_min'] != nc_rule.from_port:
            return False
        if backend_rule['port_range_max'] != nc_rule.to_port:
            return False
        if backend_rule['protocol'] != nc_rule.protocol:
            return False
        if backend_rule['remote_ip_prefix'] != nc_rule.cidr:
            return False
        return True

    def pull_flavors(self):
        nova = self.nova_admin_client
        try:
            flavors = nova.flavors.findall(is_public=True)
        except nova_exceptions.ClientException as e:
            reraise(e)

        flavor_exclude_regex = self.settings.options.get('flavor_exclude_regex', '')
        name_pattern = re.compile(flavor_exclude_regex) if flavor_exclude_regex else None
        with transaction.atomic():
            cur_flavors = self._get_current_properties(models.Flavor)
            for backend_flavor in flavors:
                if name_pattern is not None and name_pattern.match(backend_flavor.name) is not None:
                    logger.debug('Skipping pull of %s flavor as it matches %s regex pattern.',
                                 backend_flavor.name, flavor_exclude_regex)
                    continue

                cur_flavors.pop(backend_flavor.id, None)
                models.Flavor.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_flavor.id,
                    defaults={
                        'name': backend_flavor.name,
                        'cores': backend_flavor.vcpus,
                        'ram': backend_flavor.ram,
                        'disk': self.gb2mb(backend_flavor.disk),
                    })

            models.Flavor.objects.filter(backend_id__in=cur_flavors.keys()).delete()

    def pull_images(self):
        self._pull_images(models.Image, lambda image: image['visibility'] == 'public')

    @log_backend_action('push quotas for tenant')
    def push_tenant_quotas(self, tenant, quotas):
        cinder_quotas = {
            'gigabytes': self.mb2gb(quotas.get('storage')) if 'storage' in quotas else None,
            'volumes': quotas.get('volumes'),
            'snapshots': quotas.get('snapshots'),
        }
        cinder_quotas = {k: v for k, v in cinder_quotas.items() if v is not None}

        nova_quotas = {
            'instances': quotas.get('instances'),
            'cores': quotas.get('vcpu'),
            'ram': quotas.get('ram'),
        }
        nova_quotas = {k: v for k, v in nova_quotas.items() if v is not None}

        neutron_quotas = {
            'security_group': quotas.get('security_group_count'),
            'security_group_rule': quotas.get('security_group_rule_count'),
        }
        neutron_quotas = {k: v for k, v in neutron_quotas.items() if v is not None}

        try:
            if cinder_quotas:
                self.cinder_client.quotas.update(tenant.backend_id, **cinder_quotas)
            if nova_quotas:
                self.nova_client.quotas.update(tenant.backend_id, **nova_quotas)
            if neutron_quotas:
                self.neutron_client.update_quota(tenant.backend_id, {'quota': neutron_quotas})
        except Exception as e:
            reraise(e)

    @log_backend_action('pull quotas for tenant')
    def pull_tenant_quotas(self, tenant):
        self._pull_tenant_quotas(tenant.backend_id, tenant)

    def pull_floating_ips(self, tenants=None):
        if tenants is None:
            tenants = models.Tenant.objects.filter(
                state=models.Tenant.States.OK,
                service_project_link__service__settings=self.settings).prefetch_related('floating_ips')
        tenant_mappings = {tenant.backend_id: tenant for tenant in tenants}
        if not tenant_mappings:
            return

        backend_floating_ips = self.list_floatingips(tenant_mappings.keys())

        tenant_floating_ips = dict()
        for tenant_id, floating_ips in groupby(backend_floating_ips, lambda x: x['tenant_id']):
            tenant_floating_ips[tenant_mappings[tenant_id]] = floating_ips

        with transaction.atomic():
            for tenant, floating_ips in tenant_floating_ips.items():
                self._update_tenant_floating_ips(tenant, floating_ips)

            self._remove_stale_floating_ips(tenants, backend_floating_ips)

    @method_decorator(create_batch_fetcher)
    def list_floatingips(self, tenants):
        neutron = self.neutron_admin_client

        try:
            return neutron.list_floatingips(tenant_id=tenants)['floatingips']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    @log_backend_action('pull floating IPs for tenant')
    def pull_tenant_floating_ips(self, tenant):
        neutron = self.neutron_client

        try:
            backend_floating_ips = neutron.list_floatingips(tenant_id=self.tenant_id)['floatingips']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        with transaction.atomic():
            self._update_tenant_floating_ips(tenant, backend_floating_ips)
            self._remove_stale_floating_ips([tenant], backend_floating_ips)

    def _remove_stale_floating_ips(self, tenants, backend_floating_ips):
        remote_ids = {ip['id'] for ip in backend_floating_ips}
        stale_ips = models.FloatingIP.objects.filter(
            tenant__in=tenants,
            state__in=[models.FloatingIP.States.OK, models.FloatingIP.States.ERRED]
        ).exclude(backend_id__in=remote_ids)
        stale_ips.delete()

    def _update_tenant_floating_ips(self, tenant, backend_floating_ips):
        floating_ips = {ip.backend_id: ip for ip in tenant.floating_ips.filter(
            state__in=[models.FloatingIP.States.OK, models.FloatingIP.States.ERRED])}

        for backend_ip in backend_floating_ips:
            imported_floating_ip = self._backend_floating_ip_to_floating_ip(
                backend_ip,
                tenant=tenant,
                service_project_link=tenant.service_project_link,
            )
            floating_ip = floating_ips.pop(imported_floating_ip.backend_id, None)
            if floating_ip is None:
                imported_floating_ip.save()
                continue

            # Don't update user defined name.
            if floating_ip.address != floating_ip.name:
                imported_floating_ip.name = floating_ip.name
            update_pulled_fields(floating_ip, imported_floating_ip, models.FloatingIP.get_backend_fields())
            handle_resource_update_success(floating_ip)

    def _backend_floating_ip_to_floating_ip(self, backend_floating_ip, **kwargs):
        floating_ip = models.FloatingIP(
            name=backend_floating_ip['floating_ip_address'],
            description=backend_floating_ip['description'],
            address=backend_floating_ip['floating_ip_address'],
            backend_network_id=backend_floating_ip['floating_network_id'],
            runtime_state=backend_floating_ip['status'],
            backend_id=backend_floating_ip['id'],
            state=models.FloatingIP.States.OK,
        )
        for field, value in kwargs.items():
            setattr(floating_ip, field, value)

        return floating_ip

    def pull_security_groups(self, tenants=None):

        if tenants is None:
            tenants = models.Tenant.objects.filter(
                state=models.Tenant.States.OK,
                service_project_link__service__settings=self.settings,
            ).prefetch_related('security_groups')
        tenant_mappings = {tenant.backend_id: tenant for tenant in tenants}
        if not tenant_mappings:
            return

        backend_security_groups = self.list_security_groups(tenant_mappings.keys())

        tenant_security_groups = dict()
        for tenant_id, security_groups in groupby(backend_security_groups, lambda x: x['tenant_id']):
            tenant_security_groups[tenant_mappings[tenant_id]] = security_groups

        with transaction.atomic():
            for tenant, security_groups in tenant_security_groups.items():
                self._update_tenant_security_groups(tenant, security_groups)
            self._remove_stale_security_groups(tenants, backend_security_groups)

    @method_decorator(create_batch_fetcher)
    def list_security_groups(self, tenants):
        neutron = self.neutron_admin_client

        try:
            return neutron.list_security_groups(tenant_id=tenants)['security_groups']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    @log_backend_action('pull security groups for tenant')
    def pull_tenant_security_groups(self, tenant):
        neutron = self.neutron_client
        try:
            backend_security_groups = neutron.list_security_groups(tenant_id=self.tenant_id)['security_groups']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        with transaction.atomic():
            self._update_tenant_security_groups(tenant, backend_security_groups)
            self._remove_stale_security_groups([tenant], backend_security_groups)

    def _remove_stale_security_groups(self, tenants, backend_security_groups):
        remote_ids = {ip['id'] for ip in backend_security_groups}
        stale_ips = models.SecurityGroup.objects.filter(
            tenant__in=tenants,
            state__in=[models.SecurityGroup.States.OK, models.SecurityGroup.States.ERRED]
        ).exclude(backend_id__in=remote_ids)
        stale_ips.delete()

    def _update_tenant_security_groups(self, tenant, backend_security_groups):
        security_group_uuids = []
        for backend_security_group in backend_security_groups:
            imported_security_group = self._backend_security_group_to_security_group(
                backend_security_group, tenant=tenant, service_project_link=tenant.service_project_link)

            try:
                security_group = tenant.security_groups.get(backend_id=imported_security_group.backend_id)
            except models.SecurityGroup.DoesNotExist:
                imported_security_group.save()
                security_group = imported_security_group
            else:
                update_pulled_fields(security_group, imported_security_group,
                                     models.SecurityGroup.get_backend_fields())
                handle_resource_update_success(security_group)

            security_group_uuids.append(security_group.uuid)
            self._extract_security_group_rules(security_group, backend_security_group)

    def _backend_security_group_to_security_group(self, backend_security_group, **kwargs):
        security_group = models.SecurityGroup(
            name=backend_security_group['name'],
            description=backend_security_group['description'],
            backend_id=backend_security_group['id'],
            state=models.SecurityGroup.States.OK,
        )

        for field, value in kwargs.items():
            setattr(security_group, field, value)

        return security_group

    def pull_networks(self):
        tenants = models.Tenant.objects.exclude(backend_id='').filter(
            state=models.Tenant.States.OK,
            service_project_link__service__settings=self.settings,
        ).prefetch_related('networks')

        self._pull_networks(tenants)

    def _pull_tenant_networks(self, tenant):
        return self._pull_networks([tenant])

    def _pull_networks(self, tenants):
        tenant_mappings = {tenant.backend_id: tenant for tenant in tenants}
        backend_networks = self.list_networks(tenant_mappings.keys())

        networks = []
        with transaction.atomic():
            for backend_network in backend_networks:
                tenant = tenant_mappings.get(backend_network['tenant_id'])
                if not tenant:
                    logger.debug('Skipping network %s synchronization because its tenant %s is not available.',
                                 backend_network['id'], backend_network['tenant_id'])
                    continue

                imported_network = self._backend_network_to_network(
                    backend_network, tenant=tenant, service_project_link=tenant.service_project_link)

                try:
                    network = tenant.networks.get(backend_id=imported_network.backend_id)
                except models.Network.DoesNotExist:
                    imported_network.save()
                    network = imported_network
                else:
                    update_pulled_fields(network, imported_network, models.Network.get_backend_fields())
                    handle_resource_update_success(network)

                networks.append(network)

            networks_uuid = [network_item.uuid for network_item in networks]
            stale_networks = models.Network.objects.filter(
                state__in=[models.Network.States.OK, models.Network.States.ERRED],
                tenant__in=tenants).exclude(uuid__in=networks_uuid)
            stale_networks.delete()

        return networks

    @method_decorator(create_batch_fetcher)
    def list_networks(self, tenants):
        neutron = self.neutron_client
        try:
            return neutron.list_networks(tenant_id=tenants)['networks']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    def _backend_network_to_network(self, backend_network, **kwargs):
        network = models.Network(
            name=backend_network['name'],
            description=backend_network['description'],
            is_external=backend_network['router:external'],
            runtime_state=backend_network['status'],
            backend_id=backend_network['id'],
            state=models.Network.States.OK,
        )
        if backend_network.get('provider:network_type'):
            network.type = backend_network['provider:network_type']
        if backend_network.get('provider:segmentation_id'):
            network.segmentation_id = backend_network['provider:segmentation_id']

        for field, value in kwargs.items():
            setattr(network, field, value)

        return network

    def pull_subnets(self, tenant=None):
        neutron = self.neutron_client

        if tenant:
            networks = tenant.networks.all()
        else:
            networks = models.Network.objects.filter(
                state=models.Network.States.OK,
                service_project_link__service__settings=self.settings,
            )
        network_mappings = {network.backend_id: network for network in networks}
        if not network_mappings:
            return

        try:
            if tenant:
                backend_subnets = neutron.list_subnets(tenant_id=tenant.backend_id)['subnets']
            else:
                # We can't filter subnets by network IDs because it exceeds maximum request length
                backend_subnets = neutron.list_subnets()['subnets']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        subnet_uuids = []
        with transaction.atomic():
            for backend_subnet in backend_subnets:
                network = network_mappings.get(backend_subnet['network_id'])

                if not network:
                    logger.debug('OpenStack network is not imported yet. Network ID: %s',
                                 backend_subnet['network_id'])
                    continue

                imported_subnet = self._backend_subnet_to_subnet(
                    backend_subnet, network=network, service_project_link=network.service_project_link)

                try:
                    subnet = network.subnets.get(backend_id=imported_subnet.backend_id)
                except models.SubNet.DoesNotExist:
                    imported_subnet.save()
                    subnet = imported_subnet
                else:
                    update_pulled_fields(subnet, imported_subnet, models.SubNet.get_backend_fields())
                    handle_resource_update_success(subnet)

                subnet_uuids.append(subnet.uuid)

            stale_subnets = models.SubNet.objects.filter(
                state__in=[models.SubNet.States.OK, models.SubNet.States.ERRED],
                network__in=networks).exclude(uuid__in=subnet_uuids)
            stale_subnets.delete()

    @log_backend_action()
    def import_tenant_subnets(self, tenant):
        self.pull_subnets(tenant)

    def _backend_subnet_to_subnet(self, backend_subnet, **kwargs):
        subnet = models.SubNet(
            name=backend_subnet['name'],
            description=backend_subnet['description'],
            allocation_pools=backend_subnet['allocation_pools'],
            cidr=backend_subnet['cidr'],
            ip_version=backend_subnet['ip_version'],
            enable_dhcp=backend_subnet['enable_dhcp'],
            gateway_ip=backend_subnet.get('gateway_ip'),
            dns_nameservers=backend_subnet['dns_nameservers'],
            backend_id=backend_subnet['id'],
            state=models.SubNet.States.OK,
        )

        for field, value in kwargs.items():
            setattr(subnet, field, value)

        return subnet

    @log_backend_action()
    def create_tenant(self, tenant):
        keystone = self.keystone_admin_client
        try:
            backend_tenant = keystone.projects.create(
                name=tenant.name, description=tenant.description, domain=self._get_domain())
            tenant.backend_id = backend_tenant.id
            tenant.save(update_fields=['backend_id'])
        except keystone_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def create_tenant_safe(self, tenant):
        """
        Check available tenant name before creating tenant.
        It allows to avoid failure when name is already taken.
        """
        new_name = self.get_available_tenant_name(tenant.name)
        if new_name != tenant.name:
            tenant.name = new_name
            tenant.save(update_fields=['name'])
        self.create_tenant(tenant)

    def get_available_tenant_name(self, name, max_length=64):
        """
        Returns a tenant name that's free on the target deployment.
        """
        keystone = self.keystone_admin_client
        try:
            tenants = keystone.projects.list(domain=self._get_domain())
        except keystone_exceptions.ClientException as e:
            reraise(e)

        names = [tenant.name for tenant in tenants]
        new_name = name

        # If the tenant name already exists, add an underscore and a random 3
        # character alphanumeric string to the tenant name until the generated name doesn't exist.
        # Truncate original name if required, so the new name does not exceed the max_length.
        while new_name in names:
            new_name = "%s_%s" % (name, get_random_string(3))
            truncation = len(new_name) - max_length
            if truncation > 0:
                new_name = "%s_%s" % (name[:-truncation], get_random_string(3))
        return new_name

    def import_tenant(self, tenant_backend_id, service_project_link=None, save=True):
        keystone = self.keystone_admin_client
        try:
            backend_tenant = keystone.projects.get(tenant_backend_id)
        except keystone_exceptions.ClientException as e:
            reraise(e)

        tenant = models.Tenant()
        tenant.name = backend_tenant.name
        tenant.description = backend_tenant.description
        tenant.backend_id = tenant_backend_id

        if save and service_project_link:
            tenant.service_project_link = service_project_link
            tenant.state = models.Tenant.States.OK
            tenant.save()
        return tenant

    @log_backend_action()
    def pull_tenant(self, tenant):
        import_time = timezone.now()
        imported_tenant = self.import_tenant(tenant.backend_id, save=False)

        tenant.refresh_from_db()
        # if tenant was not modified in Waldur database after import.
        if tenant.modified < import_time:
            update_pulled_fields(tenant, imported_tenant, ('name', 'description'))

    @log_backend_action()
    def add_admin_user_to_tenant(self, tenant):
        """ Add user from openstack settings to new tenant """
        keystone = self.keystone_admin_client

        try:
            admin_user = keystone.users.find(name=self.settings.username)
            admin_role = keystone.roles.find(name='admin')
            try:
                keystone.roles.grant(
                    user=admin_user.id,
                    role=admin_role.id,
                    project=tenant.backend_id)
            except keystone_exceptions.Conflict:
                pass
        except keystone_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action('add user to tenant')
    def create_tenant_user(self, tenant):
        keystone = self.keystone_client

        try:
            user = keystone.users.create(
                name=tenant.user_username,
                password=tenant.user_password,
                domain=self._get_domain(),
            )
            try:
                role = keystone.roles.find(name='Member')
            except keystone_exceptions.NotFound:
                role = keystone.roles.find(name='_member_')
            keystone.roles.grant(
                user=user.id,
                role=role.id,
                project=tenant.backend_id,
            )
        except keystone_exceptions.ClientException as e:
            reraise(e)

    def create_or_update_tenant_user(self, tenant):
        keystone = self.keystone_client

        try:
            keystone_user = keystone.users.find(name=tenant.user_username)
        except keystone_exceptions.NotFound:
            self.create_tenant_user(tenant)
        except keystone_exceptions.ClientException as e:
            reraise(e)
        else:
            self.change_tenant_user_password(tenant, keystone_user)

    @log_backend_action('change password for tenant user')
    def change_tenant_user_password(self, tenant, keystone_user=None):
        keystone = self.keystone_client

        try:
            if not keystone_user:
                keystone_user = keystone.users.find(name=tenant.user_username)
            keystone.users.update(user=keystone_user, password=tenant.user_password)
        except keystone_exceptions.ClientException as e:
            reraise(e)

    def get_resources_for_import(self, resource_type=None):
        return self.get_tenants_for_import()

    def get_tenants_for_import(self):
        keystone = self.keystone_admin_client
        try:
            tenants = keystone.projects.list(domain=self._get_domain())
        except keystone_exceptions.ClientException as e:
            reraise(e)

        cur_tenants = set(models.Tenant.objects.values_list('backend_id', flat=True))

        return [{
            'backend_id': tenant.id,
            'name': tenant.name,
            'description': tenant.description,
            'type': SupportedServices.get_name_for_model(models.Tenant)
        } for tenant in tenants if tenant.id not in cur_tenants]

    def get_managed_resources(self):
        return []

    @log_backend_action()
    def delete_tenant_floating_ips(self, tenant):
        if not tenant.backend_id:
            # This method will remove all floating IPs if tenant `backend_id` is not defined.
            raise OpenStackBackendError('This method should not be called if tenant has no backend_id')

        neutron = self.neutron_admin_client

        try:
            floatingips = neutron.list_floatingips(tenant_id=tenant.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for floating_ip in floatingips.get('floatingips', []):
            self._delete_backend_floating_ip(floating_ip['id'], tenant.backend_id)

    @log_backend_action()
    def delete_tenant_ports(self, tenant):
        if not tenant.backend_id:
            # This method will remove all ports if tenant `backend_id` is not defined.
            raise OpenStackBackendError('This method should not be called if tenant has no backend_id')

        neutron = self.neutron_admin_client

        try:
            ports = neutron.list_ports(tenant_id=tenant.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for port in ports.get('ports', []):
            if 'device_id' in port and \
                    port['device_owner'] in ['network:router_interface',
                                             'network:router_interface_distributed',
                                             'network:ha_router_replicated_interface']:
                logger.info("Deleting port %s interface_router from tenant %s", port['id'], tenant.backend_id)
                try:
                    neutron.remove_interface_router(port['device_id'], {'port_id': port['id']})
                except neutron_exceptions.NotFound:
                    logger.debug("Port %s interface_router is already gone from tenant %s", port['id'],
                                 tenant.backend_id)
                except neutron_exceptions.NeutronClientException as e:
                    reraise(e)

            logger.info("Deleting port %s from tenant %s", port['id'], tenant.backend_id)
            try:
                neutron.delete_port(port['id'])
            except neutron_exceptions.NotFound:
                logger.debug("Port %s is already gone from tenant %s", port['id'], tenant.backend_id)
            except neutron_exceptions.NeutronClientException as e:
                reraise(e)

    @log_backend_action()
    def delete_tenant_routers(self, tenant):
        if not tenant.backend_id:
            # This method will remove all routers if tenant `backend_id` is not defined.
            raise OpenStackBackendError('This method should not be called if tenant has no backend_id')

        neutron = self.neutron_admin_client

        try:
            routers = neutron.list_routers(tenant_id=tenant.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for router in routers.get('routers', []):
            logger.info("Deleting router %s from tenant %s", router['id'], tenant.backend_id)
            try:
                neutron.delete_router(router['id'])
            except neutron_exceptions.NotFound:
                logger.debug("Router %s is already gone from tenant %s", router['id'], tenant.backend_id)
            except neutron_exceptions.NeutronClientException as e:
                reraise(e)

    @log_backend_action()
    def delete_tenant_networks(self, tenant):
        if not tenant.backend_id:
            # This method will remove all networks if tenant `backend_id` is not defined.
            raise OpenStackBackendError('This method should not be called if tenant has no backend_id')

        neutron = self.neutron_admin_client

        try:
            networks = neutron.list_networks(tenant_id=tenant.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for network in networks.get('networks', []):
            if network['router:external']:
                continue
            for subnet in network['subnets']:
                logger.info("Deleting subnetwork %s from tenant %s", subnet, tenant.backend_id)
                try:
                    neutron.delete_subnet(subnet)
                except neutron_exceptions.NotFound:
                    logger.info("Subnetwork %s is already gone from tenant %s", subnet, tenant.backend_id)
                except neutron_exceptions.NeutronClientException as e:
                    reraise(e)

            logger.info("Deleting network %s from tenant %s", network['id'], tenant.backend_id)
            try:
                neutron.delete_network(network['id'])
            except neutron_exceptions.NotFound:
                logger.debug("Network %s is already gone from tenant %s", network['id'], tenant.backend_id)
            except neutron_exceptions.NeutronClientException as e:
                reraise(e)

        tenant.set_quota_usage(tenant.Quotas.network_count, 0)
        tenant.set_quota_usage(tenant.Quotas.subnet_count, 0)

    @log_backend_action()
    def delete_tenant_security_groups(self, tenant):
        neutron = self.neutron_client

        try:
            sgroups = neutron.list_security_groups(tenant_id=tenant.backend_id)['security_groups']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for sgroup in sgroups:
            logger.info("Deleting security group %s from tenant %s", sgroup['id'], tenant.backend_id)
            try:
                neutron.delete_security_group(sgroup['id'])
            except neutron_exceptions.NotFound:
                logger.debug("Security group %s is already gone from tenant %s", sgroup['id'], tenant.backend_id)
            except neutron_exceptions.NeutronClientException as e:
                reraise(e)

    @log_backend_action()
    def delete_tenant_instances(self, tenant):
        nova = self.nova_client

        try:
            servers = nova.servers.list()
        except nova_exceptions.ClientException as e:
            reraise(e)

        for server in servers:
            logger.info("Deleting instance %s from tenant %s", server.id, tenant.backend_id)
            try:
                server.delete()
            except nova_exceptions.NotFound:
                logger.debug("Instance %s is already gone from tenant %s", server.id, tenant.backend_id)
            except nova_exceptions.ClientException as e:
                reraise(e)

    @log_backend_action()
    def are_all_tenant_instances_deleted(self, tenant):
        nova = self.nova_client

        try:
            servers = nova.servers.list()
        except nova_exceptions.ClientException as e:
            reraise(e)
        else:
            return not servers

    @log_backend_action()
    def delete_tenant_snapshots(self, tenant):
        cinder = self.cinder_client

        try:
            snapshots = cinder.volume_snapshots.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)

        for snapshot in snapshots:
            logger.info("Deleting snapshot %s from tenant %s", snapshot.id, tenant.backend_id)
            try:
                snapshot.delete()
            except cinder_exceptions.NotFound:
                logger.debug("Snapshot %s is already gone from tenant %s", snapshot.id, tenant.backend_id)
            except cinder_exceptions.ClientException as e:
                reraise(e)

    @log_backend_action()
    def are_all_tenant_snapshots_deleted(self, tenant):
        cinder = self.cinder_client

        try:
            snapshots = cinder.volume_snapshots.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)
        else:
            return not snapshots

    @log_backend_action()
    def delete_tenant_volumes(self, tenant):
        cinder = self.cinder_client

        try:
            volumes = cinder.volumes.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)

        for volume in volumes:
            logger.info("Deleting volume %s from tenant %s", volume.id, tenant.backend_id)
            try:
                volume.force_delete()
            except cinder_exceptions.NotFound:
                logger.debug("Volume %s is already gone from tenant %s", volume.id, tenant.backend_id)
            except cinder_exceptions.ClientException as e:
                reraise(e)

    @log_backend_action()
    def are_all_tenant_volumes_deleted(self, tenant):
        cinder = self.cinder_client

        try:
            volumes = cinder.volumes.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)
        else:
            return not volumes

    @log_backend_action()
    def delete_tenant_user(self, tenant):
        keystone = self.keystone_client
        try:
            user = keystone.users.find(name=tenant.user_username)
            logger.info('Deleting user %s that was connected to tenant %s', user.name, tenant.backend_id)
            user.delete()
        except keystone_exceptions.NotFound:
            logger.debug("User %s is already gone from tenant %s", tenant.user_username, tenant.backend_id)
        except keystone_exceptions.ClientException as e:
            logger.error('Cannot delete user %s from tenant %s. Error: %s', tenant.user_username, tenant.backend_id, e)

    @log_backend_action()
    def delete_tenant(self, tenant):
        if not tenant.backend_id:
            raise OpenStackBackendError('This method should not be called if tenant has no backend_id')

        keystone = self.keystone_admin_client

        logger.info("Deleting tenant %s", tenant.backend_id)
        try:
            keystone.projects.delete(tenant.backend_id)
        except keystone_exceptions.NotFound:
            logger.debug("Tenant %s is already gone", tenant.backend_id)
        except keystone_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def push_security_group_rules(self, security_group):
        neutron = self.neutron_client

        try:
            backend_security_group = neutron.show_security_group(security_group.backend_id)['security_group']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        backend_rules = {
            rule['id']: self._normalize_security_group_rule(rule)
            for rule in backend_security_group['security_group_rules']
            # TODO: Currently only security group rules for incoming traffic can be created
            # Therefore we should not modify security group rules for outgoing traffic
            if rule['direction'] == 'ingress'
        }

        # list of nc rules, that do not exist in openstack
        nonexistent_rules = []
        # list of nc rules, that have wrong parameters in openstack
        unsynchronized_rules = []
        # list of os rule ids, that exist in openstack and do not exist in nc
        extra_rule_ids = backend_rules.keys()

        for nc_rule in security_group.rules.all():
            if nc_rule.backend_id not in backend_rules:
                nonexistent_rules.append(nc_rule)
            else:
                backend_rule = backend_rules[nc_rule.backend_id]
                if not self._are_rules_equal(backend_rule, nc_rule):
                    unsynchronized_rules.append(nc_rule)
                extra_rule_ids.remove(nc_rule.backend_id)

        # deleting extra rules
        for backend_rule_id in extra_rule_ids:
            logger.debug('About to delete security group rule with id %s in backend', backend_rule_id)
            try:
                neutron.delete_security_group_rule(backend_rule_id)
            except neutron_exceptions.NeutronClientException:
                logger.exception('Failed to remove rule with id %s from security group %s in backend',
                                 backend_rule_id, security_group)
            else:
                logger.info('Security group rule with id %s successfully deleted in backend', backend_rule_id)

        # deleting unsynchronized rules
        for nc_rule in unsynchronized_rules:
            logger.debug('About to delete security group rule with id %s', nc_rule.backend_id)
            try:
                neutron.delete_security_group_rule(nc_rule.backend_id)
            except neutron_exceptions.NeutronClientException:
                logger.exception('Failed to remove rule with id %s from security group %s in backend',
                                 nc_rule.backend_id, security_group)
            else:
                logger.info('Security group rule with id %s successfully deleted in backend',
                            nc_rule.backend_id)

        # creating nonexistent and unsynchronized rules
        for nc_rule in unsynchronized_rules + nonexistent_rules:
            logger.debug('About to create security group rule with id %s in backend', nc_rule.id)
            try:
                # The database has empty strings instead of nulls
                if nc_rule.protocol == '':
                    nc_rule_protocol = None
                else:
                    nc_rule_protocol = nc_rule.protocol

                neutron.create_security_group_rule({'security_group_rule': {
                    'security_group_id': security_group.backend_id,
                    # XXX: Currently only security groups for incoming traffic can be created
                    'direction': 'ingress',
                    'protocol': nc_rule_protocol,
                    'port_range_min': nc_rule.from_port if nc_rule.from_port != -1 else None,
                    'port_range_max': nc_rule.to_port if nc_rule.to_port != -1 else None,
                    'remote_ip_prefix': nc_rule.cidr,
                }})
            except neutron_exceptions.NeutronClientException as e:
                logger.exception('Failed to create rule %s for security group %s in backend',
                                 nc_rule, security_group)
                reraise(e)
            else:
                logger.info('Security group rule with id %s successfully created in backend', nc_rule.id)

    @log_backend_action()
    def create_security_group(self, security_group):
        neutron = self.neutron_client
        try:
            backend_security_group = neutron.create_security_group({'security_group': {
                'name': security_group.name,
                'description': security_group.description,
            }})['security_group']
            security_group.backend_id = backend_security_group['id']
            security_group.save(update_fields=['backend_id'])
            self.push_security_group_rules(security_group)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    @log_backend_action()
    def delete_security_group(self, security_group):
        neutron = self.neutron_client
        try:
            neutron.delete_security_group(security_group.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        security_group.decrease_backend_quotas_usage()

    @log_backend_action()
    def update_security_group(self, security_group):
        neutron = self.neutron_client
        data = {'name': security_group.name, 'description': security_group.description}
        try:
            neutron.update_security_group(security_group.backend_id, {'security_group': data})
            self.push_security_group_rules(security_group)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    @log_backend_action()
    def detect_external_network(self, tenant):
        neutron = self.neutron_client
        try:
            routers = neutron.list_routers(tenant_id=tenant.backend_id)['routers']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        if bool(routers):
            router = routers[0]
        else:
            logger.warning('Tenant %s (PK: %s) does not have connected routers.', tenant, tenant.pk)
            return

        ext_gw = router.get('external_gateway_info', {})
        if ext_gw and 'network_id' in ext_gw:
            tenant.external_network_id = ext_gw['network_id']
            tenant.save()
            logger.info('Found and set external network with id %s for tenant %s (PK: %s)',
                        ext_gw['network_id'], tenant, tenant.pk)

    @log_backend_action()
    def create_network(self, network):
        neutron = self.neutron_admin_client

        data = {'name': network.name, 'tenant_id': network.tenant.backend_id}
        try:
            response = neutron.create_network({'networks': [data]})
        except neutron_exceptions.NeutronException as e:
            reraise(e)
        else:
            backend_network = response['networks'][0]
            network.backend_id = backend_network['id']
            network.runtime_state = backend_network['status']
            if backend_network.get('provider:network_type'):
                network.type = backend_network['provider:network_type']
            if backend_network.get('provider:segmentation_id'):
                network.segmentation_id = backend_network['provider:segmentation_id']
            network.save()
            # XXX: temporary fix - right now backend logic is based on statement "one tenant has one network"
            # We need to fix this in the future.
            network.tenant.internal_network_id = network.backend_id
            network.tenant.save()

    @log_backend_action()
    def update_network(self, network):
        neutron = self.neutron_admin_client

        data = {'name': network.name}
        try:
            neutron.update_network(network.backend_id, {'network': data})
        except neutron_exceptions.NeutronException as e:
            reraise(e)

    @log_backend_action()
    def delete_network(self, network):
        for subnet in network.subnets.all():
            self.delete_subnet(subnet)

        neutron = self.neutron_admin_client
        try:
            neutron.delete_network(network.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        else:
            network.decrease_backend_quotas_usage()

    @log_backend_action()
    def import_tenant_networks(self, tenant):
        networks = self._pull_tenant_networks(tenant)
        if networks:
            # XXX: temporary fix - right now backend logic is based on statement "one tenant has one network"
            # We need to fix this in the future.
            tenant.internal_network_id = networks[0].backend_id
            tenant.save()

    def import_network(self, network_backend_id):
        neutron = self.neutron_admin_client
        try:
            backend_network = neutron.show_network(network_backend_id)['network']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        return self._backend_network_to_network(backend_network)

    @log_backend_action()
    def pull_network(self, network):
        import_time = timezone.now()
        imported_network = self.import_network(network.backend_id)

        network.refresh_from_db()
        if network.modified < import_time:
            update_pulled_fields(network, imported_network, models.Network.get_backend_fields())

    @log_backend_action()
    def create_subnet(self, subnet):
        neutron = self.neutron_admin_client

        data = {
            'name': subnet.name,
            'network_id': subnet.network.backend_id,
            'tenant_id': subnet.network.tenant.backend_id,
            'cidr': subnet.cidr,
            'allocation_pools': subnet.allocation_pools,
            'ip_version': subnet.ip_version,
            'enable_dhcp': subnet.enable_dhcp,
        }
        if subnet.dns_nameservers:
            data['dns_nameservers'] = subnet.dns_nameservers
        data.update(self._serialize_subnet_gateway(subnet))
        try:
            response = neutron.create_subnet({'subnets': [data]})
            # Automatically create router for subnet
            # TODO: Ideally: Create separate model for router and create it separately.
            self.connect_router(subnet.network.name, response['subnets'][0]['id'],
                                tenant_id=subnet.network.tenant.backend_id)
        except neutron_exceptions.NeutronException as e:
            reraise(e)
        else:
            backend_subnet = response['subnets'][0]
            subnet.backend_id = backend_subnet['id']
            if backend_subnet.get('gateway_ip'):
                subnet.gateway_ip = backend_subnet['gateway_ip']
            subnet.save()

    @log_backend_action()
    def update_subnet(self, subnet):
        neutron = self.neutron_admin_client

        data = {'name': subnet.name}
        data.update(self._serialize_subnet_gateway(subnet))
        try:
            neutron.update_subnet(subnet.backend_id, {'subnet': data})
        except neutron_exceptions.NeutronException as e:
            reraise(e)

    def _serialize_subnet_gateway(self, subnet):
        data = {}
        if subnet.disable_gateway:
            data['gateway_ip'] = None
        elif subnet.gateway_ip:
            data['gateway_ip'] = subnet.gateway_ip
        return data

    @log_backend_action()
    def delete_subnet(self, subnet):
        neutron = self.neutron_admin_client
        try:
            neutron.delete_subnet(subnet.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        else:
            subnet.decrease_backend_quotas_usage()

    def import_subnet(self, subnet_backend_id):
        neutron = self.neutron_admin_client
        try:
            backend_subnet = neutron.show_subnet(subnet_backend_id)['subnet']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        subnet = models.SubNet(
            name=backend_subnet['name'],
            description=backend_subnet['description'],
            allocation_pools=backend_subnet['allocation_pools'],
            cidr=backend_subnet['cidr'],
            ip_version=backend_subnet.get('ip_version'),
            gateway_ip=backend_subnet.get('gateway_ip'),
            enable_dhcp=backend_subnet.get('enable_dhcp', False),
            state=models.Network.States.OK,
        )
        return subnet

    @log_backend_action()
    def pull_subnet(self, subnet):
        import_time = timezone.now()
        imported_subnet = self.import_subnet(subnet.backend_id)

        subnet.refresh_from_db()
        if subnet.modified < import_time:
            update_fields = ('name', 'cidr', 'allocation_pools', 'ip_version', 'gateway_ip', 'enable_dhcp')
            update_pulled_fields(subnet, imported_subnet, update_fields)

    @log_backend_action('pull floating ip')
    def pull_floating_ip(self, floating_ip):
        neutron = self.neutron_client
        try:
            backend_floating_ip = neutron.show_floatingip(floating_ip.backend_id)['floatingip']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        imported_floating_ip = self._backend_floating_ip_to_floating_ip(backend_floating_ip, tenant=floating_ip.tenant)
        update_pulled_fields(floating_ip, imported_floating_ip, models.FloatingIP.get_backend_fields())

    @log_backend_action('delete floating ip')
    def delete_floating_ip(self, floating_ip):
        self._delete_backend_floating_ip(floating_ip.backend_id, floating_ip.tenant.backend_id)
        floating_ip.decrease_backend_quotas_usage()

    @log_backend_action('create floating ip')
    def create_floating_ip(self, floating_ip):
        neutron = self.neutron_client
        try:
            backend_floating_ip = neutron.create_floatingip({
                'floatingip': {
                    'floating_network_id': floating_ip.tenant.external_network_id,
                    'tenant_id': floating_ip.tenant.backend_id,
                }
            })['floatingip']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        else:
            floating_ip.runtime_state = backend_floating_ip['status']
            floating_ip.address = backend_floating_ip['floating_ip_address']
            floating_ip.name = backend_floating_ip['floating_ip_address']
            floating_ip.backend_id = backend_floating_ip['id']
            floating_ip.backend_network_id = backend_floating_ip['floating_network_id']
            floating_ip.save()

    @log_backend_action()
    def connect_tenant_to_external_network(self, tenant, external_network_id):
        neutron = self.neutron_admin_client
        logger.debug('About to connect tenant to external network "%s" (PK: %s)', tenant.name, tenant.pk)

        try:
            # check if the network actually exists
            response = neutron.show_network(external_network_id)
        except neutron_exceptions.NeutronClientException as e:
            logger.exception('External network %s does not exist. Stale data in database?', external_network_id)
            reraise(e)

        network_name = response['network']['name']
        subnet_id = response['network']['subnets'][0]
        self.connect_router(network_name, subnet_id, external=True, network_id=response['network']['id'])

        tenant.external_network_id = external_network_id
        tenant.save()

        logger.info('Router between external network %s and tenant %s was successfully created',
                    external_network_id, tenant.backend_id)

        return external_network_id

    def _get_router(self, tenant_id):
        neutron = self.neutron_client

        try:
            routers = neutron.list_routers(tenant_id=tenant_id)['routers']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        # If any router in Tenant exists, use it
        return routers[0] if routers else None

    def _create_router(self, router_name, tenant_id):
        neutron = self.neutron_client
        create_ha_routers = bool(self.settings.options.get('create_ha_routers'))
        options = {'router': {'name': router_name, 'tenant_id': tenant_id, 'ha': create_ha_routers}}

        try:
            router = neutron.create_router(options)['router']
            logger.info('Router %s has been created.', router['name'])
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        return router

    def _connect_network_to_router(self, router, external, tenant_id, network_id=None, subnet_id=None):
        neutron = self.neutron_client
        try:
            if external:
                if (not router.get('external_gateway_info') or
                        router['external_gateway_info'].get('network_id') != network_id):
                    neutron.add_gateway_router(router['id'], {'network_id': network_id})
                    logger.info('External network %s was connected to the router %s.', network_id, router['name'])
                else:
                    logger.info('External network %s is already connected to router %s.', network_id, router['name'])
            else:
                ports = neutron.list_ports(device_id=router['id'], tenant_id=tenant_id)['ports']
                if not ports:
                    neutron.add_interface_router(router['id'], {'subnet_id': subnet_id})
                    logger.info('Internal subnet %s was connected to the router %s.', subnet_id, router['name'])
                else:
                    logger.info('Internal subnet %s is already connected to the router %s.', subnet_id, router['name'])
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

    def connect_router(self, network_name, subnet_id, external=False, network_id=None, tenant_id=None):
        tenant_id = tenant_id or self.tenant_id
        router_name = '{0}-router'.format(network_name)
        router = self._get_router(tenant_id) or self._create_router(router_name, tenant_id)
        self._connect_network_to_router(router, external, tenant_id, network_id, subnet_id)

        return router['id']

    @log_backend_action()
    def update_tenant(self, tenant):
        keystone = self.keystone_admin_client
        try:
            keystone.projects.update(tenant.backend_id, name=tenant.name, description=tenant.description)
        except keystone_exceptions.NotFound as e:
            logger.error('Tenant with id %s does not exist', tenant.backend_id)
            reraise(e)

    def pull_service_settings_quotas(self):
        nova = self.nova_admin_client
        try:
            stats = nova.hypervisor_stats.statistics()
        except nova_exceptions.ClientException as e:
            reraise(e)

        self.settings.set_quota_limit(self.settings.Quotas.openstack_vcpu, stats.vcpus)
        self.settings.set_quota_usage(self.settings.Quotas.openstack_vcpu, stats.vcpus_used)

        self.settings.set_quota_limit(self.settings.Quotas.openstack_ram, stats.memory_mb)
        self.settings.set_quota_usage(self.settings.Quotas.openstack_ram, stats.memory_mb_used)

        self.settings.set_quota_usage(self.settings.Quotas.openstack_storage, self.get_storage_usage())

    def get_storage_usage(self):
        cinder = self.cinder_admin_client

        try:
            volumes = cinder.volumes.list()
            snapshots = cinder.volume_snapshots.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)

        storage = sum(self.gb2mb(v.size) for v in volumes + snapshots)
        return storage

    def get_stats(self):
        tenants = models.Tenant.objects.filter(service_project_link__service__settings=self.settings)
        quota_names = ('vcpu', 'ram', 'storage')
        quota_values = models.Tenant.get_sum_of_quotas_as_dict(
            tenants, quota_names=quota_names, fields=['limit'])
        quota_stats = {
            'vcpu_quota': quota_values.get('vcpu', -1.0),
            'ram_quota': quota_values.get('ram', -1.0),
            'storage_quota': quota_values.get('storage', -1.0)
        }

        stats = {}
        for quota in self.settings.quotas.all():
            name = quota.name.replace('openstack_', '')
            if name not in quota_names:
                continue
            stats[name] = quota.limit
            stats[name + '_usage'] = quota.usage
        stats.update(quota_stats)
        return stats



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0007_add_instance_tenant'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instance',
            name='tenant',
            field=models.ForeignKey(related_name='instances', to='openstack.Tenant'),
        ),
    ]



from __future__ import unicode_literals

from waldur_core.cost_tracking import views


def register_in(router):
    router.register(r'price-estimates', views.PriceEstimateViewSet)
    router.register(r'default-price-list-items', views.DefaultPriceListItemViewSet)
    router.register(r'service-price-list-items', views.PriceListItemViewSet)
    router.register(r'merged-price-list-items', views.MergedPriceListItemViewSet, base_name='merged-price-list-item')



from django.contrib.contenttypes.models import ContentType

from waldur_core.core.utils import DryRunCommand
from waldur_mastermind.invoices.models import InvoiceItem
from waldur_mastermind.marketplace.models import Resource
from waldur_mastermind.packages.models import OpenStackPackage


class Command(DryRunCommand):
    help = """Import OpenStack invoice items from packages application."""

    def handle(self, dry_run, *args, **options):
        ct = ContentType.objects.get_for_model(OpenStackPackage)
        for invoice_item in InvoiceItem.objects.filter(content_type=ct).exclude(object_id=None):
            package = invoice_item.scope
            if not package:
                continue
            tenant = package.tenant
            try:
                resource = Resource.objects.get(scope=tenant)
            except Resource.DoesNotExist:
                self.stdout.write(self.style.ERROR('Marketplace resource for tenant with ID %s is not found.') %
                                  invoice_item.scope.pk)
            else:
                if dry_run:
                    self.stdout.write(self.style.SUCCESS('Importing invoice item for package with ID %s.') %
                                      invoice_item.scope.pk)
                else:
                    invoice_item.scope = resource
                    invoice_item.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-08 06:56
from __future__ import unicode_literals

from django.db import migrations, models

from ..utils import fill_activated_field


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0081_allow_ico_mime_type_for_images'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderitem',
            name='activated',
            field=models.DateTimeField(blank=True, null=True, verbose_name='activation date'),
        ),
        migrations.AlterField(
            model_name='offeringcomponent',
            name='billing_type',
            field=models.CharField(choices=[('fixed', 'Fixed-price'), ('usage', 'Usage-based'), ('one', 'One-time'), ('few', 'One-time on plan switch')], default='fixed', max_length=5),
        ),
        migrations.RunPython(fill_activated_field, elidable=True),
    ]



from __future__ import unicode_literals

from django.contrib import admin
from django.conf.urls import url
from django.core.urlresolvers import resolve
from django.forms.models import ModelForm
from django.http import HttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _, ungettext
from waldur_core.core.admin import JsonWidget

from waldur_core.core import admin as core_admin
from waldur_core.core.admin import format_json_field
from waldur_core.core.admin_filters import RelatedOnlyDropdownFilter
from waldur_core.structure.models import ServiceSettings, SharedServiceSettings, PrivateServiceSettings

from . import models, tasks


class ServiceProviderAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'customer', 'created')


class AttributeOptionInline(admin.TabularInline):
    model = models.AttributeOption


class AttributeAdmin(admin.ModelAdmin):
    inlines = [AttributeOptionInline]
    list_display = ('title', 'get_category', 'section', 'type', 'key', 'required',)
    list_filter = ('section',)
    ordering = ('section', 'title')

    def get_category(self, obj):
        return obj.section.category

    get_category.short_description = _('Category')
    get_category.admin_order_field = 'section__category__title'


class AttributeInline(admin.TabularInline):
    model = models.Attribute


class SectionAdmin(admin.ModelAdmin):
    inlines = [AttributeInline]
    list_display = ('title', 'category', 'key')


class SectionInline(admin.TabularInline):
    model = models.Section


class CategoryColumnInline(admin.TabularInline):
    model = models.CategoryColumn
    list_display = ('index', 'title', 'attribute', 'widget')


class CategoryComponentInline(admin.TabularInline):
    model = models.CategoryComponent


class CategoryAdmin(admin.ModelAdmin):
    list_display = ('title', 'uuid',)
    inlines = [SectionInline, CategoryColumnInline, CategoryComponentInline]


class ScreenshotsInline(admin.StackedInline):
    model = models.Screenshot
    fields = ('name', 'description', 'image')


class PlansInline(admin.StackedInline):
    model = models.Plan
    fields = ('name', 'description', 'unit_price', 'unit',
              'product_code', 'article_code', 'archived', 'max_amount')


class ConnectedResourceMixin(object):
    """
    Protects object from modification if there are connected resources.
    """

    protected_fields = ()

    def get_readonly_fields(self, request, obj=None):
        fields = super(ConnectedResourceMixin, self).get_readonly_fields(request, obj)
        if obj and obj.has_connected_resources:
            return fields + self.protected_fields
        else:
            return fields

    def has_delete_permission(self, request, obj=None):
        if obj and obj.has_connected_resources:
            return False
        return True


class ParentInlineMixin(object):
    def get_parent_object_from_request(self, request):
        """
        Returns the parent object from the request or None.

        Note that this only works for Inlines, because the `parent_model`
        is not available in the regular admin.ModelAdmin as an attribute.
        """
        resolved = resolve(request.path_info)
        if resolved.args:
            return self.parent_model.objects.get(pk=resolved.args[0])
        return None


class PlanComponentInline(ConnectedResourceMixin,
                          ParentInlineMixin,
                          admin.TabularInline):
    model = models.PlanComponent
    protected_fields = ('component', 'amount', 'price')

    def has_add_permission(self, request, obj=None):
        plan = self.get_parent_object_from_request(request)
        if plan and plan.has_connected_resources:
            return False
        else:
            return True

    def get_extra(self, request, obj=None, **kwargs):
        plan = self.get_parent_object_from_request(request)
        if plan and plan.has_connected_resources:
            return 0
        else:
            return super(PlanComponentInline, self).get_extra(request, obj, **kwargs)


class PlanAdmin(ConnectedResourceMixin, admin.ModelAdmin):
    list_display = ('name', 'offering', 'archived', 'unit', 'unit_price')
    list_filter = ('offering', 'archived')
    search_fields = ('name', 'offering__name')
    inlines = [PlanComponentInline]
    protected_fields = ('unit', 'unit_price', 'product_code', 'article_code')
    readonly_fields = ('scope_link', 'backend_id')
    fields = (
        'name', 'description', 'unit', 'unit_price',
        'product_code', 'article_code', 'max_amount',
        'archived',
    ) + readonly_fields

    def scope_link(self, obj):
        return get_admin_link_for_scope(obj.scope)

    scope_link.short_description = 'Scope'


class OfferingAdminForm(ModelForm):
    class Meta:
        widgets = {
            'attributes': JsonWidget(),
            'geolocations': JsonWidget(),
            'options': JsonWidget(),
        }


class OfferingComponentInline(admin.StackedInline):
    model = models.OfferingComponent


def get_admin_url_for_scope(scope):
    if isinstance(scope, ServiceSettings):
        model = scope.shared and SharedServiceSettings or PrivateServiceSettings
    else:
        model = scope
    return reverse('admin:%s_%s_change' % (scope._meta.app_label, model._meta.model_name), args=[scope.id])


def get_admin_link_for_scope(scope):
    return format_html('<a href="{}">{}</a>', get_admin_url_for_scope(scope), scope)


class OfferingAdmin(admin.ModelAdmin):
    form = OfferingAdminForm
    inlines = [ScreenshotsInline, PlansInline, OfferingComponentInline]
    list_display = ('name', 'customer', 'state', 'category', 'billable')
    list_filter = ('state', 'shared', 'billable', ('category', RelatedOnlyDropdownFilter),)
    search_fields = ('name', 'uuid')
    fields = ('state', 'customer', 'category', 'name', 'native_name',
              'description', 'native_description', 'full_description',
              'rating', 'thumbnail', 'attributes', 'options', 'geolocations',
              'shared', 'billable', 'allowed_customers', 'type', 'scope_link', 'vendor_details',
              'paused_reason')
    readonly_fields = ('rating', 'scope_link')

    def scope_link(self, obj):
        if obj.scope:
            return format_html('<a href="{}">{}</a>', get_admin_url_for_scope(obj.scope), obj.scope)

    actions = ['activate']

    def activate(self, request, queryset):
        valid_states = [models.Offering.States.DRAFT, models.Offering.States.PAUSED]
        valid_offerings = queryset.filter(state__in=valid_states)
        count = valid_offerings.count()

        for offering in valid_offerings:
            offering.activate()
            offering.save()

        message = ungettext(
            'One offering has been activated.',
            '%(count)d offerings have been activated.',
            count
        )
        message = message % {'count': count}

        self.message_user(request, message)

    activate.short_description = _('Activate offerings')


class OrderItemInline(admin.TabularInline):
    model = models.OrderItem
    fields = ('offering', 'state', 'attributes', 'cost', 'plan')
    readonly_fields = fields


class OrderAdmin(core_admin.ReadOnlyAdminMixin, core_admin.ExtraActionsMixin, admin.ModelAdmin):
    list_display = ('uuid', 'project', 'created', 'created_by', 'state', 'total_cost')
    fields = ['created_by', 'approved_by', 'approved_at', 'created', 'project', 'state',
              'total_cost', 'modified', 'pdf_file']
    readonly_fields = ('created', 'modified', 'created_by', 'approved_by', 'approved_at',
                       'project', 'approved_by', 'total_cost', 'pdf_file')

    list_filter = ('state', 'created')
    ordering = ('-created',)
    inlines = [OrderItemInline]

    def get_extra_actions(self):
        return [
            self.create_pdf_for_all,
        ]

    def get_urls(self):
        my_urls = [
            url(r'^(.+)/change/pdf_file/$', self.admin_site.admin_view(self.pdf_file_view)),
        ]
        return my_urls + super(OrderAdmin, self).get_urls()

    def create_pdf_for_all(self, request):
        tasks.create_pdf_for_all.delay()
        message = _('PDF creation has been scheduled')
        self.message_user(request, message)
        return redirect(reverse('admin:marketplace_order_changelist'))

    def pdf_file_view(self, request, pk=None):
        order = models.Order.objects.get(id=pk)
        file_response = HttpResponse(order.file, content_type='application/pdf')
        filename = order.get_filename()
        file_response['Content-Disposition'] = 'attachment; filename="{filename}"'.format(filename=filename)
        return file_response

    def pdf_file(self, obj):
        if not obj.file:
            return ''

        return format_html('<a href="./pdf_file">download</a>')

    create_pdf_for_all.name = _('Create PDF for all orders')


class ResourceForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ResourceForm, self).__init__(*args, **kwargs)
        if self.instance and hasattr(self.instance, 'offering'):
            # Filter marketplace resource plans by offering
            self.fields['plan'].queryset = self.fields['plan'].queryset.filter(
                offering=self.instance.offering)


class ResourceAdmin(admin.ModelAdmin):
    form = ResourceForm
    list_display = ('name', 'project', 'state', 'category', 'created')
    list_filter = (
        'state',
        ('project', RelatedOnlyDropdownFilter),
        ('offering', RelatedOnlyDropdownFilter),
    )
    readonly_fields = ('state', 'scope_link', 'project_link', 'offering_link',
                                'plan_link', 'formatted_attributes', 'formatted_limits')
    fields = readonly_fields + ('plan',)
    date_hierarchy = 'created'
    search_fields = ('name', 'uuid')

    def category(self, obj):
        return obj.offering.category

    def scope_link(self, obj):
        return get_admin_link_for_scope(obj.scope)

    scope_link.short_description = 'Scope'

    def project_link(self, obj):
        return get_admin_link_for_scope(obj.project)

    project_link.short_description = 'Project'

    def offering_link(self, obj):
        return get_admin_link_for_scope(obj.offering)

    offering_link.short_description = 'Offering'

    def plan_link(self, obj):
        return get_admin_link_for_scope(obj.plan)

    plan_link.short_description = 'Plan'

    def formatted_attributes(self, obj):
        return format_json_field(obj.attributes)

    formatted_attributes.allow_tags = True
    formatted_attributes.short_description = 'Attributes'

    def formatted_limits(self, obj):
        return format_json_field(obj.limits)

    formatted_limits.allow_tags = True
    formatted_limits.short_description = 'Limits'


admin.site.register(models.ServiceProvider, ServiceProviderAdmin)
admin.site.register(models.Category, CategoryAdmin)
admin.site.register(models.Offering, OfferingAdmin)
admin.site.register(models.Section, SectionAdmin)
admin.site.register(models.Attribute, AttributeAdmin)
admin.site.register(models.Screenshot)
admin.site.register(models.Order, OrderAdmin)
admin.site.register(models.Plan, PlanAdmin)
admin.site.register(models.Resource, ResourceAdmin)



from . import views


def register_in(router):
    router.register(r'zabbix', views.ZabbixServiceViewSet, base_name='zabbix')
    router.register(r'zabbix-hosts', views.HostViewSet, base_name='zabbix-host')
    router.register(r'zabbix-itservices', views.ITServiceViewSet, base_name='zabbix-itservice')
    router.register(r'zabbix-service-project-link', views.ZabbixServiceProjectLinkViewSet, base_name='zabbix-spl')
    router.register(r'zabbix-templates', views.TemplateViewSet, base_name='zabbix-template')
    router.register(r'zabbix-triggers', views.TriggerViewSet, base_name='zabbix-trigger')
    router.register(r'zabbix-user-groups', views.UserGroupViewSet, base_name='zabbix-user-group')
    router.register(r'zabbix-users', views.UserViewSet, base_name='zabbix-user')



import logging

from django.conf import settings
from django_fsm import TransitionNotAllowed
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, exceptions, status, response, views

from waldur_core.core import views as core_views
from waldur_core.structure import permissions as structure_permissions

from . import backend, filters, log, models, serializers


logger = logging.getLogger(__name__)


class ExtensionDisabled(exceptions.APIException):
    status_code = status.HTTP_424_FAILED_DEPENDENCY
    default_detail = _('PayPal extension is disabled.')


class CheckExtensionMixin(object):
    """ Raise exception if paypal extension is disabled """

    def initial(self, request, *args, **kwargs):
        if not settings.WALDUR_PAYPAL['ENABLED']:
            raise ExtensionDisabled()
        return super(CheckExtensionMixin, self).initial(request, *args, **kwargs)


class CreateByStaffOrOwnerMixin(object):

    def create(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        customer = serializer.validated_data['customer']
        if not structure_permissions._has_owner_access(request.user, customer):
            raise exceptions.PermissionDenied()

        return super(CreateByStaffOrOwnerMixin, self).create(request)


class PaymentView(CheckExtensionMixin, CreateByStaffOrOwnerMixin, core_views.ProtectedViewSet):
    queryset = models.Payment.objects.all()
    serializer_class = serializers.PaymentSerializer
    lookup_field = 'uuid'
    filter_class = filters.PaymentFilter

    def perform_create(self, serializer):
        """
        Create new payment via Paypal gateway
        """

        return_url = serializer.validated_data.pop('return_url')
        cancel_url = serializer.validated_data.pop('cancel_url')

        payment = serializer.save()

        try:
            backend_payment = payment.get_backend().make_payment(
                payment.amount, payment.tax,
                description='Replenish account in Waldur for %s' % payment.customer.name,
                return_url=return_url,
                cancel_url=cancel_url)

            payment.backend_id = backend_payment.payment_id
            payment.approval_url = backend_payment.approval_url
            payment.token = backend_payment.token
            payment.set_created()
            payment.save()

            serializer.instance = payment

            log.event_logger.paypal_payment.info(
                'Created new payment for {customer_name}',
                event_type='payment_creation_succeeded',
                event_context={'payment': payment}
            )

        except backend.PayPalError as e:
            message = 'Unable to create payment because of backend error %s' % e
            logger.warning(message)
            payment.set_erred()
            payment.error_message = message
            payment.save()
            raise exceptions.APIException()

    def get_payment(self, token):
        """
        Find Paypal payment object in the database by token
        and check if current user has access to it.
        :param token: string
        :return: Payment object
        """
        error_message = "Payment with token %s does not exist" % token

        try:
            payment = models.Payment.objects.get(token=token)
        except models.Payment.DoesNotExist:
            raise exceptions.NotFound(error_message)

        if not structure_permissions._has_owner_access(self.request.user, payment.customer):
            raise exceptions.NotFound(error_message)

        return payment

    @decorators.list_route(methods=['POST'])
    def approve(self, request):
        """
        Approve Paypal payment.
        """
        serializer = serializers.PaymentApproveSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        payment_id = serializer.validated_data['payment_id']
        payer_id = serializer.validated_data['payer_id']
        token = serializer.validated_data['token']
        payment = self.get_payment(token)

        try:
            payment.get_backend().approve_payment(payment_id, payer_id)

            payment.set_approved()
            payment.error_message = ''
            payment.save()

            log.event_logger.paypal_payment.info(
                'Payment for {customer_name} has been approved.',
                event_type='payment_approval_succeeded',
                event_context={'payment': payment}
            )
            return response.Response({'detail': 'Payment has been approved.'}, status=status.HTTP_200_OK)

        except backend.PayPalError as e:
            message = 'Unable to approve payment because of backend error %s' % e
            logger.warning(message)
            payment.error_message = message
            payment.save()
            raise exceptions.APIException(message)

        except TransitionNotAllowed:
            message = 'Unable to approve payment because of invalid state.'
            payment.set_erred()
            payment.error_message = message
            payment.save()
            return response.Response({'detail': message}, status=status.HTTP_409_CONFLICT)

    @decorators.list_route(methods=['POST'])
    def cancel(self, request):
        """
        Cancel Paypal payment.
        """
        serializer = serializers.PaymentCancelSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        token = serializer.validated_data['token']
        payment = self.get_payment(token)

        try:
            payment.set_cancelled()
            payment.save()

            log.event_logger.paypal_payment.info(
                'Payment for {customer_name} has been cancelled.',
                event_type='payment_cancel_succeeded',
                event_context={'payment': payment}
            )
            return response.Response({'detail': 'Payment has been cancelled.'}, status=status.HTTP_200_OK)

        except TransitionNotAllowed:
            return response.Response({'detail': 'Unable to cancel payment because of invalid state.'},
                                     status=status.HTTP_409_CONFLICT)


class InvoicesViewSet(CheckExtensionMixin, core_views.ReadOnlyActionsViewSet):
    queryset = models.Invoice.objects.all()
    serializer_class = serializers.InvoiceSerializer
    lookup_field = 'uuid'
    filter_class = filters.InvoiceFilter


class InvoiceWebHookViewSet(CheckExtensionMixin, views.APIView):
    authentication_classes = ()
    permission_classes = ()
    serializer_class = serializers.InvoiceUpdateWebHookSerializer
    valid_event_types = [
        'INVOICING.INVOICE.CANCELLED',
        'INVOICING.INVOICE.PAID',
        'INVOICING.INVOICE.REFUNDED',
        'INVOICING.INVOICE.UPDATED',
        # 'INVOICING.INVOICE.CREATED', No need to update created Invoice
    ]

    def post(self, request, *args, **kwargs):
        if request.data.get('event_type') not in self.valid_event_types:
            return response.Response(status=status.HTTP_304_NOT_MODIFIED)

        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response(status=status.HTTP_200_OK)



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers
from waldur_core.structure.models import VirtualMachine


from . import models, validators, exceptions


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': _('Rancher server URL'),
        'username': _('Rancher access key'),
        'password': _('Rancher secret key'),
    }

    SERVICE_ACCOUNT_EXTRA_FIELDS = {}

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.RancherService
        required_fields = ('backend_url', 'username', 'password')


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.RancherServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'vmware-detail'},
        }


class ClusterSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='rancher-detail',
        read_only=True,
        lookup_field='uuid',
    )

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rancher-spl-detail',
        queryset=models.RancherServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    instance = core_serializers.GenericRelatedField(
        related_models=VirtualMachine.get_all_models(),
        required=True,
        write_only=True,
    )

    name = serializers.CharField(max_length=150, validators=[validators.ClusterNameValidator])

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Cluster
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'instance', 'node_command',
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'instance',
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'node_command',
        )

        extra_kwargs = dict(
            cluster={
                'view_name': 'rancher-cluster-detail',
                'lookup_field': 'uuid',
            },
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def create(self, validated_data):
        instance = validated_data.pop('instance')
        cluster = super(ClusterSerializer, self).create(validated_data)
        models.Node.objects.create(instance=instance, cluster=cluster)
        return cluster

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        instance = attrs.get('instance')

        if models.Node.objects.filter(
                object_id=instance.id,
                content_type=ContentType.objects.get_for_model(instance)
        ).exists():
            raise serializers.ValidationError({'instance': 'The selected instance is already in use.'})

        return super(ClusterSerializer, self).validate(attrs)


class NodeSerializer(serializers.HyperlinkedModelSerializer):
    instance = core_serializers.GenericRelatedField(
        related_models=VirtualMachine.get_all_models(),
        required=True,
    )

    class Meta(object):
        model = models.Node
        fields = ('uuid', 'url', 'created', 'modified', 'cluster', 'instance')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'rancher-node-detail'},
            'cluster': {'lookup_field': 'uuid', 'view_name': 'rancher-cluster-detail'}
        }

    def validate(self, attrs):
        instance = attrs.get('instance')

        if models.Node.objects.filter(
                object_id=instance.id,
                content_type=ContentType.objects.get_for_model(instance)
        ).exists():
            raise serializers.ValidationError({'instance': 'The selected instance is already in use.'})

        return super(NodeSerializer, self).validate(attrs)


class ClusterImportableSerializer(serializers.Serializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rancher-spl-detail',
        queryset=models.RancherServiceProjectLink.objects.all(),
        write_only=True)

    name = serializers.CharField(read_only=True)
    backend_id = serializers.CharField(source="id", read_only=True)
    kubernetes_version = serializers.CharField(
        source="rancherKubernetesEngineConfig.kubernetesVersion",
        read_only=True)
    created_ts = serializers.IntegerField(read_only=True)
    nodes = serializers.ListField(
        source="appliedSpec.rancherKubernetesEngineConfig.nodes",
        required=False,
        read_only=True)


class ClusterImportSerializer(ClusterImportableSerializer):
    backend_id = serializers.CharField(write_only=True)

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Cluster.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({'backend_id': _('Cluster has been imported already.')})

        try:
            backend = service_project_link.get_backend()
            cluster = backend.import_cluster(backend_id, service_project_link=service_project_link)
        except exceptions.RancherException:
            raise serializers.ValidationError({
                'backend_id': _("Can't import cluster with ID %s") % validated_data['backend_id']
            })

        return cluster



from __future__ import unicode_literals

from waldur_mastermind.booking import views


def register_in(router):
    router.register(r'booking-resources', views.ResourceViewSet,
                    base_name='booking-resource')



from __future__ import unicode_literals

from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status
from rest_framework import serializers as rf_serializers
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from waldur_core.core import validators as core_validators
from waldur_core.core import views as core_views
from waldur_mastermind.marketplace import filters as marketplace_filters
from waldur_mastermind.marketplace import models

from . import PLUGIN_NAME, filters, serializers
from .log import event_logger


class ResourceViewSet(core_views.ReadOnlyActionsViewSet):
    queryset = models.Resource.objects.filter(offering__type=PLUGIN_NAME)
    filter_backends = (
        DjangoFilterBackend,
        filters.OfferingCustomersFilterBackend,
    )
    filter_class = marketplace_filters.ResourceFilter
    lookup_field = 'uuid'
    serializer_class = serializers.BookingResourceSerializer

    @detail_route(methods=['post'])
    def reject(self, request, uuid=None):
        resource = self.get_object()

        with transaction.atomic():
            try:
                order_item = models.OrderItem.objects.get(
                    resource=resource,
                    offering=resource.offering,
                    type=models.OrderItem.Types.CREATE,
                    state=models.OrderItem.States.EXECUTING,
                )
            except models.OrderItem.DoesNotExist:
                raise rf_serializers.ValidationError(_('Resource rejecting is not available because '
                                                       'the reference order item is not found.'))
            except models.OrderItem.MultipleObjectsReturned:
                raise rf_serializers.ValidationError(_('Resource rejecting is not available because '
                                                       'several reference order items are found.'))
            order_item.set_state_terminated()
            order_item.save()
            resource.set_state_terminated()
            resource.save()

        return Response({'order_item_uuid': order_item.uuid}, status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def accept(self, request, uuid=None):
        resource = self.get_object()

        with transaction.atomic():
            try:
                order_item = models.OrderItem.objects.get(
                    resource=resource,
                    offering=resource.offering,
                    type=models.OrderItem.Types.CREATE,
                    state=models.OrderItem.States.EXECUTING,
                )
            except models.OrderItem.DoesNotExist:
                raise rf_serializers.ValidationError(_('Resource accepting is not available because '
                                                       'the reference order item is not found.'))
            except models.OrderItem.MultipleObjectsReturned:
                raise rf_serializers.ValidationError(_('Resource accepting is not available because '
                                                       'several reference order items are found.'))
            order_item.set_state_done()
            order_item.save()
            resource.set_state_ok()
            resource.save()

            event_logger.waldur_booking.info(
                'Device booking {resource_name} has been accepted.',
                event_type='device_booking_is_accepted',
                event_context={
                    'resource': resource,
                })

        return Response({'order_item_uuid': order_item.uuid}, status=status.HTTP_200_OK)

    reject_validators = accept_validators = [
        core_validators.StateValidator(models.Resource.States.CREATING)
    ]



from __future__ import unicode_literals

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from ... import models, executors, handlers


class Command(BaseCommand):
    help_text = "Add default security groups with given names to all tenants."

    def add_arguments(self, parser):
        parser.add_argument('names', nargs='+', type=str)

    def handle(self, *args, **options):
        names = options['names']
        default_security_groups = getattr(settings, 'WALDUR_OPENSTACK', {}).get('DEFAULT_SECURITY_GROUPS')
        security_groups = []
        for name in names:
            try:
                group = next(sg for sg in default_security_groups if sg['name'] == name)
            except StopIteration:
                raise CommandError('There is no default security group with name %s' % name)
            else:
                security_groups.append(group)

        for tenant in models.Tenant.objects.all():
            for group in security_groups:
                if tenant.security_groups.filter(name=group['name']).exists():
                    self.stdout.write('Tenant %s already has security group %s' % (tenant, group['name']))
                    continue
                tenant.security_groups.create(name=group['name'],
                                              description=group['description'],
                                              service_project_link=tenant.service_project_link,
                                              )
                try:
                    db_security_group = handlers.create_security_group(tenant, group)
                except handlers.SecurityGroupCreateException as e:
                    self.stdout.write(
                        'Failed to add security_group %s to tenant %s. Error: %s' % (group['name'], tenant, e))
                else:
                    try:
                        executors.SecurityGroupCreateExecutor.execute(db_security_group, async=False)
                    except Exception as e:
                        self.stdout.write('Failed to add security group %s to tenant %s. Error: %s' %
                                          (db_security_group, tenant, e))
                    else:
                        self.stdout.write(
                            'Security group %s has been successfully added to tenant %s' % (group['name'], tenant))



from __future__ import unicode_literals

from django.contrib import admin

from waldur_core.structure import admin as structure_admin

from . import models


admin.site.register(models.RancherService, structure_admin.ServiceAdmin)
admin.site.register(models.RancherServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(models.Cluster)
admin.site.register(models.Node)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-19 08:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import taggit.managers
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers
import waldur_core.structure.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('taggit', '0002_auto_20150616_2121'),
    ]

    operations = [
        migrations.CreateModel(
            name='Cluster',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True)),
                ('node_command', models.CharField(blank=True, help_text='Rancher generated node installation command base.', max_length=1024)),
            ],
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Node',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('cluster', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_rancher.Cluster')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='RancherService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'verbose_name': 'Rancher provider',
                'verbose_name_plural': 'Rancher providers',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='RancherServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_rancher.RancherService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Rancher provider project link',
                'verbose_name_plural': 'Rancher provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='rancherservice',
            name='projects',
            field=models.ManyToManyField(related_name='rancher_services', through='waldur_rancher.RancherServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='rancherservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='cluster',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='k8s_clusters', to='waldur_rancher.RancherServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='cluster',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AlterUniqueTogether(
            name='rancherserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='rancherservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='node',
            unique_together=set([('content_type', 'object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='cluster',
            unique_together=set([('service_project_link', 'name'), ('service_project_link', 'backend_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-08-21 09:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0090_resource_current_usages'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='paused_reason',
            field=models.TextField(blank=True),
        ),
    ]



import decimal
import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db import transaction

from waldur_core.structure import models as structure_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.packages import models as package_models
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack.apps import OpenStackConfig
from waldur_openstack.openstack_tenant import apps as openstack_tenant_apps
from waldur_openstack.openstack_tenant import models as openstack_tenant_models

from . import INSTANCE_TYPE, PACKAGE_TYPE, VOLUME_TYPE, RAM_TYPE, STORAGE_TYPE, utils

logger = logging.getLogger(__name__)


def create_template_for_plan(sender, instance, created=False, **kwargs):
    plan = instance

    if plan.scope:
        return

    if not created:
        return

    if plan.offering.type != PACKAGE_TYPE:
        return

    if not isinstance(plan.offering.scope, structure_models.ServiceSettings):
        logger.warning('Skipping plan synchronization because offering scope is not service settings. '
                       'Plan ID: %s', plan.id)
        return

    if plan.offering.scope.type != OpenStackConfig.service_name:
        logger.warning('Skipping plan synchronization because service settings type is not OpenStack. '
                       'Plan ID: %s', plan.id)
        return

    with transaction.atomic():
        template = package_models.PackageTemplate.objects.create(
            service_settings=plan.offering.scope,
            name=plan.name,
            description=plan.description,
            product_code=plan.product_code,
            article_code=plan.article_code,
            unit=plan.unit,
        )
        plan.scope = template
        plan.save()


PLAN_FIELDS = {'name', 'archived', 'product_code', 'article_code'}


def update_template_for_plan(sender, instance, created=False, **kwargs):
    plan = instance

    if plan.offering.type != PACKAGE_TYPE:
        return

    if created:
        return

    update_fields = set(plan.tracker.changed()) & PLAN_FIELDS
    if not update_fields:
        return

    if not plan.scope:
        return

    template = plan.scope
    for field in update_fields:
        setattr(template, field, getattr(plan, field))
    template.save(update_fields=update_fields)


def update_plan_for_template(sender, instance, created=False, **kwargs):
    template = instance

    if created:
        return

    update_fields = set(template.tracker.changed()) & PLAN_FIELDS
    if not update_fields:
        return

    try:
        plan = marketplace_models.Plan.objects.get(scope=template)
    except (ObjectDoesNotExist, MultipleObjectsReturned):
        return

    for field in update_fields:
        setattr(plan, field, getattr(template, field))
    plan.save(update_fields=update_fields)


def synchronize_plan_component(sender, instance, created=False, **kwargs):
    component = instance

    if not created and not set(instance.tracker.changed()) & {'amount', 'price'}:
        return

    if component.plan.offering.type != PACKAGE_TYPE:
        return

    template = component.plan.scope
    if not template:
        logger.warning('Skipping plan component synchronization because offering does not have scope. '
                       'Offering ID: %s', component.plan.offering.id)
        return

    amount = component.amount
    price = component.price

    # In marketplace RAM and storage is stored in GB, but in package plugin it is stored in MB.
    if component.component.type in (RAM_TYPE, STORAGE_TYPE):
        amount = amount * 1024
        price = decimal.Decimal(price) / decimal.Decimal(1024.0)

    package_component = package_models.PackageComponent.objects.filter(
        template=template, type=component.component.type).first()

    if package_component:
        package_component.amount = amount
        package_component.price = price
        package_component.save(update_fields=['amount', 'price'])

    elif created:
        package_models.PackageComponent.objects.create(
            template=template,
            type=component.component.type,
            amount=amount,
            price=price,
        )


def create_offering_from_tenant(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    if instance.tracker.previous('state') != instance.States.CREATING:
        return

    if instance.state != instance.States.OK:
        return

    create_offerings_for_volume_and_instance(instance)


def create_offerings_for_volume_and_instance(tenant):
    if not settings.WALDUR_MARKETPLACE_OPENSTACK['AUTOMATICALLY_CREATE_PRIVATE_OFFERING']:
        return

    try:
        resource = marketplace_models.Resource.objects.get(scope=tenant)
    except ObjectDoesNotExist:
        logger.debug('Skipping offering creation for tenant because order '
                     'item does not exist. OpenStack tenant ID: %s', tenant.id)
        return

    try:
        service_settings = structure_models.ServiceSettings.objects.get(
            scope=tenant,
            type=openstack_tenant_apps.OpenStackTenantConfig.service_name,
        )
    except ObjectDoesNotExist:
        logger.debug('Skipping offering creation for tenant because service settings '
                     'object does not exist. OpenStack tenant ID: %s', tenant.id)
        return

    parent_offering = resource.offering
    for offering_type in (INSTANCE_TYPE, VOLUME_TYPE):
        try:
            category, offering_name = utils.get_category_and_name_for_offering_type(
                offering_type, service_settings)
        except ObjectDoesNotExist:
            logger.warning('Skipping offering creation for tenant because category '
                           'for instances and volumes is not yet defined.')
            continue
        payload = dict(
            type=offering_type,
            name=offering_name,
            scope=service_settings,
            shared=False,
            category=category,
            # OpenStack instance and volume offerings are charged as a part of its tenant
            billable=False,
        )

        fields = (
            'state',
            'customer',
            'attributes',
            'thumbnail',
            'vendor_details',
            'geolocations',
        )
        for field in fields:
            payload[field] = getattr(parent_offering, field)

        with transaction.atomic():
            offering = marketplace_models.Offering.objects.create(**payload)
            offering.allowed_customers.add(tenant.service_project_link.project.customer)


def archive_offering(sender, instance, **kwargs):
    service_settings = instance

    if service_settings.type == openstack_tenant_apps.OpenStackTenantConfig.service_name and \
            service_settings.content_type == ContentType.objects.get_for_model(openstack_models.Tenant):
        marketplace_models.Offering.objects.filter(scope=service_settings).update(
            state=marketplace_models.Offering.States.ARCHIVED
        )


def synchronize_volume_metadata(sender, instance, created=False, **kwargs):
    volume = instance
    if not created and not set(volume.tracker.changed()) & {'size', 'instance_id'}:
        return

    try:
        resource = marketplace_models.Resource.objects.get(scope=volume)
    except ObjectDoesNotExist:
        logger.debug('Skipping resource synchronization for OpenStack volume '
                     'because marketplace resource does not exist. '
                     'Resource ID: %s', instance.id)
        return

    utils.import_volume_metadata(resource)


def synchronize_instance_metadata(sender, instance, created=False, **kwargs):
    if not created and not set(instance.tracker.changed()) & {'name'}:
        return

    for volume in instance.volumes.all():
        try:
            resource = marketplace_models.Resource.objects.get(scope=volume)
        except ObjectDoesNotExist:
            logger.debug('Skipping resource synchronization for OpenStack volume '
                         'because marketplace resource does not exist. '
                         'Resource ID: %s', instance.id)
            continue

        resource.backend_metadata['instance_name'] = volume.instance.name
        resource.save(update_fields=['backend_metadata'])


def synchronize_internal_ips(sender, instance, created=False, **kwargs):
    internal_ip = instance
    if not created and not set(internal_ip.tracker.changed()) & {'ip4_address', 'instance_id'}:
        return

    vms = {vm for vm in (internal_ip.instance_id, internal_ip.tracker.previous('instance_id')) if vm}

    for vm in vms:
        try:
            scope = openstack_tenant_models.Instance.objects.get(id=vm)
            resource = marketplace_models.Resource.objects.get(scope=scope)
        except ObjectDoesNotExist:
            logger.debug('Skipping resource synchronization for OpenStack instance '
                         'because marketplace resource does not exist. '
                         'Resource ID: %s', vm)
            continue

        utils.import_instance_metadata(resource)


def synchronize_floating_ips(sender, instance, created=False, **kwargs):
    floating_ip = instance
    if not created and not set(instance.tracker.changed()) & {'address', 'internal_ip_id'}:
        return

    internal_ips = {ip for ip in (floating_ip.internal_ip_id, floating_ip.tracker.previous('internal_ip_id')) if ip}
    for ip_id in internal_ips:
        try:
            scope = openstack_tenant_models.Instance.objects.get(internal_ips_set__id=ip_id)
            resource = marketplace_models.Resource.objects.get(scope=scope)
        except ObjectDoesNotExist:
            logger.debug('Skipping resource synchronization for OpenStack instance '
                         'because marketplace resource does not exist. '
                         'Resource ID: %s', ip_id)
            continue

        utils.import_instance_metadata(resource)


def create_resource_of_volume_if_instance_created(sender, instance, created=False, **kwargs):
    resource = instance

    if not created or not resource.scope or not resource.offering.scope:
        return

    if resource.offering.type != INSTANCE_TYPE:
        return

    instance = resource.scope

    volume_offering = utils.get_offering(VOLUME_TYPE, resource.offering.scope)
    if not volume_offering:
        return

    for volume in instance.volumes.all():
        if marketplace_models.Resource.objects.filter(scope=volume).exists():
            continue

        volume_resource = marketplace_models.Resource(
            project=resource.project,
            offering=volume_offering,
            created=resource.created,
            name=volume.name,
            scope=volume,
        )

        volume_resource.init_cost()
        volume_resource.save()
        utils.import_volume_metadata(volume_resource)
        volume_resource.init_quotas()


def create_marketplace_resource_for_imported_resources(sender, instance, offering=None, plan=None, **kwargs):
    resource = marketplace_models.Resource(
        project=instance.service_project_link.project,
        state=utils.get_resource_state(instance.state),
        name=instance.name,
        scope=instance,
        created=instance.created,
        plan=plan,
        offering=offering,
    )

    if isinstance(instance, openstack_tenant_models.Instance):
        offering = offering or utils.get_offering(INSTANCE_TYPE, instance.service_settings)

        if not offering:
            return

        resource.offering = offering

        resource.init_cost()
        resource.save()
        utils.import_instance_metadata(resource)
        resource.init_quotas()

    if isinstance(instance, openstack_tenant_models.Volume):
        offering = offering or utils.get_offering(VOLUME_TYPE, instance.service_settings)

        if not offering:
            return

        resource.offering = offering

        resource.init_cost()
        resource.save()
        utils.import_volume_metadata(resource)
        resource.init_quotas()

    if isinstance(instance, openstack_models.Tenant):
        offering = offering or utils.get_offering(PACKAGE_TYPE, instance.service_settings)

        if not offering:
            return

        resource.offering = offering

        resource.init_cost()
        resource.save()
        utils.import_resource_metadata(resource)
        resource.init_quotas()
        create_offerings_for_volume_and_instance(instance)


def import_resource_metadata_when_resource_is_created(sender, instance, created=False, **kwargs):
    if not created:
        return

    if isinstance(instance.scope, openstack_tenant_models.Volume):
        utils.import_volume_metadata(instance)

    if isinstance(instance.scope, openstack_tenant_models.Instance):
        utils.import_instance_metadata(instance)


def update_openstack_tenant_usages(sender, instance, created=False, **kwargs):
    if created:
        return

    if not isinstance(instance.scope, openstack_models.Tenant):
        return

    tenant = instance.scope

    try:
        resource = marketplace_models.Resource.objects.get(scope=tenant)
    except ObjectDoesNotExist:
        logger.debug('Skipping usages synchronization for tenant because '
                     'resource does not exist. OpenStack tenant ID: %s', tenant.id)
        return

    utils.import_usage(resource)


def update_invoice_when_resource_is_created(sender, instance, **kwargs):
    if not settings.WALDUR_MARKETPLACE_OPENSTACK['BILLING_ENABLED']:
        return

    if instance.offering.type == PACKAGE_TYPE:
        registrators.RegistrationManager.register(instance)


def update_invoice_when_resource_is_updated(sender, order_item, **kwargs):
    if not settings.WALDUR_MARKETPLACE_OPENSTACK['BILLING_ENABLED']:
        return

    if order_item.offering.type == PACKAGE_TYPE:
        registrators.RegistrationManager.terminate(order_item.resource)
        registrators.RegistrationManager.register(order_item.resource)


def update_invoice_when_resource_is_deleted(sender, instance, **kwargs):
    if not settings.WALDUR_MARKETPLACE_OPENSTACK['BILLING_ENABLED']:
        return

    if instance.offering.type == PACKAGE_TYPE:
        registrators.RegistrationManager.terminate(instance)



def update_vm_total_disk_when_disk_is_created_or_updated(sender, instance, created=False, **kwargs):
    vm = instance.vm
    vm.disk = vm.total_disk
    vm.save(update_fields=['disk'])


def update_vm_total_disk_when_disk_is_deleted(sender, instance, **kwargs):
    vm = instance.vm
    vm.disk = vm.total_disk
    vm.save(update_fields=['disk'])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-08-01 17:04
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0020_protect_offering_issue'),
    ]

    operations = [
        migrations.AddField(
            model_name='supportuser',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active'),
        ),
        migrations.AlterField(
            model_name='issue',
            name='assignee',
            field=models.ForeignKey(blank=True, help_text='Help desk user who will implement the issue', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='issues', to='support.SupportUser'),
        ),
        migrations.AlterField(
            model_name='issue',
            name='caller',
            field=models.ForeignKey(help_text='Waldur user who has reported the issue.', on_delete=django.db.models.deletion.PROTECT, related_name='created_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='issue',
            name='reporter',
            field=models.ForeignKey(blank=True, help_text='Help desk user who have created the issue that is reported by caller.', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='reported_issues', to='support.SupportUser'),
        ),
    ]



from waldur_core.core import WaldurExtension


class PlaybookJobsExtension(WaldurExtension):
    class Settings:
        WALDUR_PLAYBOOK_JOBS = {
            'PLAYBOOKS_DIR_NAME': 'ansible_playbooks',
            'PLAYBOOK_ICON_SIZE': (64, 64),
        }

    @staticmethod
    def django_app():
        return 'waldur_ansible.playbook_jobs'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def is_assembly():
        return True



from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from libcloud.compute.types import NodeState
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers

from . import models
from .backend import AWSBackendError


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'username': '',
        'token': '',
    }

    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'images_regex': ''
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.AWSService
        extra_field_options = {
            'username': {
                'label': 'Access key ID',
                'required': True
            },
            'token': {
                'label': 'Secret access key',
                'required': True
            },
            'images_regex': {
                'help_text': _('Regular expression to limit images list')
            }
        }


class RegionSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Region
        fields = ('url', 'uuid', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'}
        }


class ImageSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Image
        fields = ('url', 'uuid', 'name', 'region')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'}
        }

    region = RegionSerializer(read_only=True)


class SizeSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Size
        fields = ('url', 'uuid', 'name', 'cores', 'ram', 'disk', 'regions', 'description')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'}
        }

    # AWS expose a more technical backend_id as a name. AWS's short codes are more popular
    name = serializers.ReadOnlyField(source='backend_id')
    description = serializers.ReadOnlyField(source='name')
    regions = RegionSerializer(many=True, read_only=True)


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.AWSServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'aws-detail'},
        }


class AWSImportSerializerMixin(object):
    def get_fields(self):
        from waldur_core.structure import SupportedServices
        fields = super(AWSImportSerializerMixin, self).get_fields()
        resources = SupportedServices.get_service_resources(models.AWSService)
        choices = [SupportedServices.get_name_for_model(resource) for resource in resources]
        fields['type'] = serializers.ChoiceField(choices=choices, write_only=True)
        return fields


class InstanceSerializer(structure_serializers.VirtualMachineSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='aws-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='aws-spl-detail',
        queryset=models.AWSServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    region = serializers.HyperlinkedRelatedField(
        view_name='aws-region-detail',
        lookup_field='uuid',
        queryset=models.Region.objects.all(),
        write_only=True)

    image = serializers.HyperlinkedRelatedField(
        view_name='aws-image-detail',
        lookup_field='uuid',
        queryset=models.Image.objects.all(),
        write_only=True)

    size = serializers.HyperlinkedRelatedField(
        view_name='aws-size-detail',
        lookup_field='uuid',
        queryset=models.Size.objects.all(),
        write_only=True)

    class Meta(structure_serializers.VirtualMachineSerializer.Meta):
        model = models.Instance
        fields = structure_serializers.VirtualMachineSerializer.Meta.fields + (
            'region', 'image', 'size'
        )
        protected_fields = structure_serializers.VirtualMachineSerializer.Meta.protected_fields + (
            'region', 'image', 'size'
        )

    def validate(self, attrs):
        attrs = super(InstanceSerializer, self).validate(attrs)

        region = attrs['region']
        image = attrs['image']
        size = attrs['size']

        if image.region != region:
            raise serializers.ValidationError(_('Image is missing in region %s') % region.name)

        if not size.regions.filter(pk=region.pk).exists():
            raise serializers.ValidationError(_('Size is missing in region %s') % region.name)

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        size = validated_data.get('size')
        ssh_key = validated_data.get('ssh_public_key')

        validated_data['ram'] = size.ram
        validated_data['cores'] = size.cores
        validated_data['disk'] = size.disk
        validated_data['size_backend_id'] = size.backend_id

        if ssh_key is not None:
            validated_data['key_name'] = ssh_key.name
            validated_data['key_fingerprint'] = ssh_key.fingerprint

        instance = super(InstanceSerializer, self).create(validated_data)
        volume = {
            'name': ('temp-%s' % instance.name)[:150],
            'state': models.Volume.States.CREATION_SCHEDULED,
            'instance': instance,
            'service_project_link': instance.service_project_link,
            'region': instance.region,
            # Size will be received from the backend
            'size': 0
        }
        models.Volume.objects.create(**volume)

        return instance


class InstanceImportSerializer(AWSImportSerializerMixin,
                               structure_serializers.BaseResourceImportSerializer):
    class Meta(structure_serializers.BaseResourceImportSerializer.Meta):
        model = models.Instance

    def create(self, validated_data):
        backend = self.context['service'].get_backend()
        try:
            region, instance = backend.find_instance(validated_data['backend_id'])
        except AWSBackendError:
            raise serializers.ValidationError(
                {'backend_id': _("Can't find instance with ID %s") % validated_data['backend_id']})

        validated_data['name'] = instance['name']
        validated_data['public_ips'] = instance['public_ips']
        validated_data['cores'] = instance['cores']
        validated_data['ram'] = instance['ram']
        validated_data['disk'] = instance['disk']
        validated_data['created'] = instance['created']
        validated_data['state'] = instance['state']
        validated_data['region'] = region
        validated_data.pop('type')

        return super(InstanceImportSerializer, self).create(validated_data)


class InstanceResizeSerializer(structure_serializers.PermissionFieldFilteringMixin,
                               serializers.Serializer):
    size = serializers.HyperlinkedRelatedField(
        view_name='aws-size-detail',
        lookup_field='uuid',
        queryset=models.Size.objects.all(),
    )

    def get_fields(self):
        fields = super(InstanceResizeSerializer, self).get_fields()
        if self.instance:
            fields['size'].query_params = {
                'region_uuid': self.instance.region.uuid
            }
        return fields

    def get_filtered_field_names(self):
        return ('size',)

    def validate(self, attrs):
        size = attrs['size']
        instance = self.instance

        if not size.regions.filter(uuid=self.instance.region.uuid).exists():
            raise serializers.ValidationError(_('New size is not within the same region.'))

        if (size.ram, size.disk, size.cores) == (self.instance.ram, self.instance.disk, self.instance.cores):
            raise serializers.ValidationError(_('New size is the same as current.'))

        if size.disk < self.instance.disk:
            raise serializers.ValidationError(_('New disk size should be greater than the previous value'))

        if instance.runtime_state not in [NodeState.TERMINATED,
                                          NodeState.STOPPED,
                                          NodeState.SUSPENDED,
                                          NodeState.PAUSED]:
            raise serializers.ValidationError(_('Instance runtime state must be in one of offline states.'))

        return attrs

    def update(self, instance, validated_data):
        size = validated_data.get('size')

        ram_increase = size.ram - instance.ram
        cores_increase = size.cores - instance.cores
        disk_increase = size.disk - instance.disk

        instance.ram = size.ram
        instance.cores = size.cores
        instance.disk = size.disk
        instance.save(update_fields=['ram', 'cores', 'disk'])

        spl = instance.service_project_link
        spl.add_quota_usage(spl.Quotas.storage, disk_increase, validate=True)
        spl.add_quota_usage(spl.Quotas.ram, ram_increase, validate=True)
        spl.add_quota_usage(spl.Quotas.vcpu, cores_increase, validate=True)

        return instance


class VolumeSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='aws-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='aws-spl-detail',
        queryset=models.AWSServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    region = serializers.HyperlinkedRelatedField(
        view_name='aws-region-detail',
        lookup_field='uuid',
        queryset=models.Region.objects.all(),
        write_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Volume
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'size', 'region', 'volume_type'
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'device', 'instance', 'runtime_state'
        )
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'size', 'region', 'volume_type', 'device', 'instance', 'runtime_state'
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'instance': {'lookup_field': 'uuid', 'view_name': 'aws-instance-detail'}
        }


class VolumeImportSerializer(AWSImportSerializerMixin,
                             structure_serializers.BaseResourceImportSerializer):
    class Meta(structure_serializers.BaseResourceImportSerializer.Meta):
        model = models.Volume

    def create(self, validated_data):
        backend = self.context['service'].get_backend()
        try:
            region, volume = backend.find_volume(validated_data['backend_id'])
        except AWSBackendError:
            raise serializers.ValidationError(
                {'backend_id': _("Can't find volume with ID %s") % validated_data['backend_id']})

        instance_id = volume['instance_id']
        if instance_id:
            try:
                instance = models.Instance.objects.get(backend_id=instance_id)
            except models.Instance.DoesNotExist:
                raise serializers.ValidationError(
                    _('You must import instance with ID %s first') % instance_id)
            else:
                validated_data['instance'] = instance

        validated_data['name'] = volume['name']
        validated_data['size'] = volume['size']
        validated_data['created'] = volume['created']
        validated_data['runtime_state'] = volume['runtime_state']
        validated_data['state'] = models.Volume.States.OK
        validated_data['device'] = volume['device']
        validated_data['volume_type'] = volume['volume_type']
        validated_data['region'] = region
        validated_data.pop('type')

        return super(VolumeImportSerializer, self).create(validated_data)


class VolumeAttachSerializer(structure_serializers.PermissionFieldFilteringMixin,
                             serializers.Serializer):
    instance = serializers.HyperlinkedRelatedField(
        view_name='aws-instance-detail',
        lookup_field='uuid',
        queryset=models.Instance.objects.all(),
    )
    device = serializers.CharField(
        max_length=128,
        help_text=_('The device name for attachment. For example, use /dev/sd[f-p] for Linux instances.')
    )

    def get_fields(self):
        fields = super(VolumeAttachSerializer, self).get_fields()
        if self.instance:
            fields['instance'].query_params = {
                'region_uuid': self.instance.region.uuid
            }
        return fields

    def get_filtered_field_names(self):
        return ('instance',)

    def validate(self, attrs):
        volume = self.instance
        instance = attrs['instance']

        if volume.instance:
            raise serializers.ValidationError(_('Volume is already attached to instance.'))

        if volume.region != instance.region:
            raise serializers.ValidationError(_('Instance is not within the same region.'))

        if instance.runtime_state not in [NodeState.TERMINATED,
                                          NodeState.STOPPED,
                                          NodeState.SUSPENDED,
                                          NodeState.PAUSED]:
            raise serializers.ValidationError(_('Instance runtime state must be in one of offline states.'))

        return attrs

    def update(self, volume, validated_data):
        volume.instance = validated_data.get('instance')
        volume.device = validated_data.get('device')
        volume.save(update_fields=['instance', 'device'])

        return volume



import copy

from django.conf import settings
from django.test import override_settings


def override_marketplace_settings(**kwargs):
    marketplace_settings = copy.deepcopy(settings.WALDUR_MARKETPLACE)
    marketplace_settings.update(kwargs)
    return override_settings(WALDUR_MARKETPLACE=marketplace_settings)



""" Defines how to optimize DigitalOcean droplets sizes """
import collections

from rest_framework import serializers as rf_serializers

from waldur_digitalocean import (
    apps as do_apps, models as do_models, serializers as do_serializers, cost_tracking as do_cost_tracking)

from .. import optimizers, register, serializers
from . import utils


OptimizedPreset = collections.namedtuple('OptimizedPreset', ('preset', 'size', 'quantity', 'price'))

OptimizedDigitalOcean = optimizers.namedtuple_with_defaults(
    'OptimizedDigitalOcean',
    field_names=optimizers.OptimizedService._fields + ('optimized_presets',),
    default_values=optimizers.OptimizedService._defaults,
)


class DigitalOceanOptimizer(optimizers.Optimizer):
    """ Find the cheapest Digital Ocean size for each preset """
    HOURS_IN_DAY = 24
    DAYS_IN_MONTH = 30

    def _get_size_prices(self, sizes, service):
        """ Return dictionary with items <size>: <size price> """
        service_price_list_items = utils.get_service_price_list_items(service, do_models.Droplet)
        size_prices = {item.key: item.value for item in service_price_list_items
                       if item.item_type == do_cost_tracking.DropletStrategy.Types.FLAVOR}
        return {size: size_prices.get(size.name, size.price) * self.HOURS_IN_DAY for size in sizes}

    def optimize(self, deployment_plan, service):
        optimized_presets = []
        price = 0
        sizes = do_models.Size.objects.all()
        size_prices = self._get_size_prices(sizes, service)
        for item in deployment_plan.items.all():
            preset = item.preset
            sizes = [size for size in sizes
                     if size.cores >= preset.cores and size.ram >= preset.ram and size.disk >= preset.storage]
            if not sizes:
                preset_as_str = '%s (cores: %s, ram %s MB, storage %s MB)' % (
                    preset.name, preset.cores, preset.ram, preset.storage)
                raise optimizers.OptimizationError(
                    'It is impossible to create a droplet for preset %s. It is too big.' % preset_as_str)
            optimal_size = min(sizes, key=lambda size: size_prices[size])
            optimized_presets.append(OptimizedPreset(
                preset=preset,
                size=optimal_size,
                quantity=item.quantity,
                price=size_prices[optimal_size] * item.quantity,
            ))
            price += size_prices[optimal_size] * item.quantity
        return OptimizedDigitalOcean(price=price, service=service, optimized_presets=optimized_presets)


register.Register.register_optimizer(do_apps.DigitalOceanConfig.service_name, DigitalOceanOptimizer)


class OptimizedPresetSerializer(rf_serializers.Serializer):
    size = do_serializers.SizeSerializer()
    preset = serializers.PresetSerializer()
    quantity = rf_serializers.IntegerField()
    price = rf_serializers.DecimalField(max_digits=22, decimal_places=10)


class OptimizedDigitalOceanSerializer(serializers.OptimizedServiceSerializer):
    service = rf_serializers.HyperlinkedRelatedField(
        view_name='digitalocean-detail',
        lookup_field='uuid',
        read_only=True,
    )
    optimized_presets = OptimizedPresetSerializer(many=True)


register.Register.register_serializer(do_apps.DigitalOceanConfig.service_name, OptimizedDigitalOceanSerializer)



from __future__ import unicode_literals

import logging
from django.contrib.contenttypes.models import ContentType

from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace import utils as marketplace_utils
from waldur_mastermind.support import models as support_models

from . import utils

logger = logging.getLogger(__name__)

BillingTypes = marketplace_models.OfferingComponent.BillingTypes
OrderTypes = marketplace_models.OrderItem.Types


class OfferingRegistrator(registrators.BaseRegistrator):
    def get_sources(self, customer):
        return support_models.Offering.objects.filter(
            project__customer=customer,
            state=support_models.Offering.States.OK,
        ).distinct()

    def get_customer(self, source):
        return source.project.customer

    def _find_item(self, source, now):
        offering = source
        result = utils.get_offering_items().filter(
            object_id=offering.id,
            invoice__customer=offering.project.customer,
            invoice__state=invoice_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        )
        return list(result)

    def _create_item(self, source, invoice, start, end, **kwargs):
        offering = source

        try:
            resource = marketplace_models.Resource.objects.get(scope=offering)
            plan = resource.plan

            if not plan:
                logger.warning('Skipping support invoice creation because '
                               'billing is not enabled for offering. '
                               'Offering ID: %s', offering.id)
                return

            self.create_items_for_plan(invoice, plan, offering, start, end, **kwargs)

        except marketplace_models.Resource.DoesNotExist:
            # If an offering isn't request based support offering
            item = invoice_models.InvoiceItem.objects.create(
                content_type=ContentType.objects.get_for_model(offering),
                object_id=offering.id,
                project=offering.project,
                invoice=invoice,
                start=start,
                end=end,
                details=self.get_details(offering),
                unit_price=offering.unit_price,
                unit=offering.unit,
                product_code=offering.product_code,
                article_code=offering.article_code
            )
            self.init_details(item)
            return item

    def create_items_for_plan(self, invoice, plan, offering, start, end, **kwargs):
        order_type = kwargs.get('order_type')

        for plan_component in plan.components.all():
            offering_component = plan_component.component

            is_fixed = offering_component.billing_type == BillingTypes.FIXED
            is_one = offering_component.billing_type == BillingTypes.ONE_TIME
            is_switch = offering_component.billing_type == BillingTypes.ON_PLAN_SWITCH

            if is_fixed or \
                    (is_one and order_type == OrderTypes.CREATE) or \
                    (is_switch and order_type == OrderTypes.UPDATE):
                details = self.get_component_details(offering, plan_component)

                unit_price = plan_component.price
                unit = plan.unit

                if is_fixed:
                    unit_price *= plan_component.amount
                elif is_one or is_switch:
                    unit = invoice_models.Units.QUANTITY

                item = invoice_models.InvoiceItem.objects.create(
                    content_type=ContentType.objects.get_for_model(offering),
                    object_id=offering.id,
                    project=offering.project,
                    invoice=invoice,
                    start=start,
                    end=end,
                    details=details,
                    unit_price=unit_price,
                    unit=unit,
                    quantity=1 if is_one or is_switch else 0,
                    product_code=offering_component.product_code or plan.product_code,
                    article_code=offering_component.article_code or plan.article_code,
                )
                self.init_details(item)

    def get_details(self, source):
        offering = source
        details = {
            'offering_type': offering.type,
            'offering_name': offering.name,
            'offering_uuid': offering.uuid.hex,
            'plan_name': offering.plan.name if offering.plan else ''
        }
        service_provider_info = marketplace_utils.get_service_provider_info(source)
        details.update(service_provider_info)
        return details

    def get_name(self, offering):
        if offering.plan:
            return '%s (%s / %s)' % (offering.name, offering.type, offering.plan.name)
        else:
            return '%s (%s)' % (offering.name, offering.type)

    def terminate(self, source, now=None):
        super(OfferingRegistrator, self).terminate(source, now)
        offering = source

        if not utils.is_request_based(offering):
            utils.get_offering_items().filter(object_id=offering.id).update(object_id=None)

    def get_component_details(self, offering, plan_component):
        details = self.get_details(offering)
        details.update({
            'plan_component_id': plan_component.id,
            'offering_component_type': plan_component.component.type,
            'offering_component_name': plan_component.component.name,
            'offering_component_measured_unit': plan_component.component.measured_unit,
        })
        return details



from __future__ import unicode_literals

import re

from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers

from . import models
from .backend import DigitalOceanBackendError


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'token': '',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.DigitalOceanService
        extra_field_options = {
            'token': {
                'label': 'Access token'
            }
        }


class RegionSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Region
        fields = ('url', 'uuid', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class ImageSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Image
        fields = ('url', 'uuid', 'name', 'distribution', 'type', 'regions',
                  'is_official', 'created_at', 'min_disk_size')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    regions = RegionSerializer(many=True, read_only=True)


class SizeSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Size
        fields = ('url', 'uuid', 'name', 'cores', 'ram', 'disk', 'transfer', 'regions')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    regions = RegionSerializer(many=True, read_only=True)


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.DigitalOceanServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'digitalocean-detail'},
        }


class DropletSerializer(structure_serializers.VirtualMachineSerializer):

    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='digitalocean-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='digitalocean-spl-detail',
        queryset=models.DigitalOceanServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    region = serializers.HyperlinkedRelatedField(
        view_name='digitalocean-region-detail',
        lookup_field='uuid',
        queryset=models.Region.objects.all(),
        write_only=True)

    image = serializers.HyperlinkedRelatedField(
        view_name='digitalocean-image-detail',
        lookup_field='uuid',
        queryset=models.Image.objects.all(),
        write_only=True)

    size = serializers.HyperlinkedRelatedField(
        view_name='digitalocean-size-detail',
        lookup_field='uuid',
        queryset=models.Size.objects.all(),
        write_only=True)

    class Meta(structure_serializers.VirtualMachineSerializer.Meta):
        model = models.Droplet
        fields = structure_serializers.VirtualMachineSerializer.Meta.fields + (
            'region', 'image', 'size', 'runtime_state', 'region_name',
        )
        protected_fields = structure_serializers.VirtualMachineSerializer.Meta.protected_fields + (
            'region', 'image', 'size',
        )
        read_only_fields = structure_serializers.VirtualMachineSerializer.Meta.read_only_fields + (
            'runtime_state', 'region_name',
        )

    def validate(self, attrs):
        attrs = super(DropletSerializer, self).validate(attrs)

        if not self.instance:
            region = attrs['region']
            image = attrs['image']
            size = attrs['size']

            if not re.match(r'[a-zA-Z0-9.-]+$', attrs['name']):
                raise serializers.ValidationError({
                    'name': _('Only valid hostname characters are allowed. (a-z, A-Z, 0-9, . and -)')
                })

            if not attrs.get('ssh_public_key') and image.is_ssh_key_mandatory:
                raise serializers.ValidationError({
                    'ssh_public_key': _('SSH public key is required for this image')
                })

            if not image.regions.filter(pk=region.pk).exists():
                raise serializers.ValidationError({
                    'image': _('Image is missing in region %s') % region
                })

            if not size.regions.filter(pk=region.pk).exists():
                raise serializers.ValidationError({
                    'size': _('Size is missing in region %s') % region
                })

            if image.min_disk_size and size.disk < image.min_disk_size:
                raise serializers.ValidationError({
                    'size': _('Disk provided by size %(size)s is not enough for image %(image)s') % ({
                        'size': size,
                        'image': image
                    })
                })

        return attrs

    def create(self, validated_data):
        validated_data['region_name'] = validated_data['region'].name
        validated_data['size_name'] = validated_data['size'].name
        return super(DropletSerializer, self).create(validated_data)


class DropletImportSerializer(structure_serializers.BaseResourceImportSerializer):

    class Meta(structure_serializers.BaseResourceImportSerializer.Meta):
        model = models.Droplet

    def create(self, validated_data):
        backend = self.context['service'].get_backend()
        backend_id = validated_data['backend_id']
        service_project_link = validated_data['service_project_link']
        try:
            return backend.import_droplet(backend_id, service_project_link)
        except DigitalOceanBackendError:
            raise serializers.ValidationError(
                {'backend_id': _("Can't find droplet with ID %s") % backend_id})


class DropletResizeSerializer(serializers.Serializer):
    size = serializers.HyperlinkedRelatedField(
        view_name='digitalocean-size-detail',
        lookup_field='uuid',
        queryset=models.Size.objects.all(),
        write_only=True)
    disk = serializers.BooleanField(required=True)

    def get_fields(self):
        fields = super(DropletResizeSerializer, self).get_fields()
        field = fields['size']
        field.value_field = 'url'
        field.display_name_field = 'name'
        return fields

    class Meta:
        fields = 'size', 'disk'

    def validate_size(self, value):
        if value:
            if self.is_same_size(value):
                raise ValidationError(_('New size is the same. Please select another one.'))

            if value.disk < self.instance.disk:
                raise ValidationError(_('Disk sizes are not allowed to be decreased through a resize operation.'))
        return value

    def is_same_size(self, new_size):
        return (new_size.disk == self.instance.disk and
                new_size.cores == self.instance.cores and
                new_size.ram == self.instance.ram)



from . import views, models


def register_in(router):
    router.register(r'python-management', views.PythonManagementViewSet, base_name=models.PythonManagement.get_url_name())
    router.register(r'pip-packages', views.PipPackagesViewSet, base_name='pip_packages')



from collections import OrderedDict

GUEST_OS_CHOICES = OrderedDict((
    ('DOS', 'MS-DOS'),
    ('WIN_31', 'Windows 3.1'),
    ('WIN_95', 'Windows 95'),
    ('WIN_98', 'Windows 98'),
    ('WIN_ME', 'Windows Millennium Edition'),
    ('WIN_NT', 'Windows NT 4'),
    ('WIN_2000_PRO', 'Windows 2000 Professional'),
    ('WIN_2000_SERV', 'Windows 2000 Server'),
    ('WIN_2000_ADV_SERV', 'Windows 2000 Advanced Server'),
    ('WIN_XP_HOME', 'Windows XP Home Edition'),
    ('WIN_XP_PRO', 'Windows XP Professional'),
    ('WIN_XP_PRO_64', 'Windows XP Professional Edition (64 bit)'),
    ('WIN_NET_WEB', 'Windows Server 2003, Web Edition'),
    ('WIN_NET_STANDARD', 'Windows Server 2003, Standard Edition'),
    ('WIN_NET_ENTERPRISE', 'Windows Server 2003, Enterprise Edition'),
    ('WIN_NET_DATACENTER', 'Windows Server 2003, Datacenter Edition'),
    ('WIN_NET_BUSINESS', 'Windows Small Business Server 2003'),
    ('WIN_NET_STANDARD_64', 'Windows Server 2003, Standard Edition (64 bit)'),
    ('WIN_NET_ENTERPRISE_64', 'Windows Server 2003, Enterprise Edition (64 bit)'),
    ('WIN_LONGHORN', 'Windows Longhorn (experimental)'),
    ('WIN_LONGHORN_64', 'Windows Longhorn (64 bit) (experimental)'),
    ('WIN_NET_DATACENTER_64', 'Windows Server 2003, Datacenter Edition (64 bit) (experimental)'),
    ('WIN_VISTA', 'Windows Vista'),
    ('WIN_VISTA_64', 'Windows Vista (64 bit)'),
    ('WINDOWS_7', 'Windows 7'),
    ('WINDOWS_7_64', 'Windows 7 (64 bit)'),
    ('WINDOWS_7_SERVER_64', 'Windows Server 2008 R2 (64 bit)'),
    ('WINDOWS_8', 'Windows 8'),
    ('WINDOWS_8_64', 'Windows 8 (64 bit)'),
    ('WINDOWS_8_SERVER_64', 'Windows 8 Server (64 bit)'),
    ('WINDOWS_9', 'Windows 10'),
    ('WINDOWS_9_64', 'Windows 10 (64 bit)'),
    ('WINDOWS_9_SERVER_64', 'Windows 10 Server (64 bit)'),
    ('WINDOWS_HYPERV', 'Windows Hyper-V'),
    ('FREEBSD', 'FreeBSD'),
    ('FREEBSD_64', 'FreeBSD x64'),
    ('REDHAT', 'Red Hat Linux 2.1'),
    ('RHEL_2', 'Red Hat Enterprise Linux 2'),
    ('RHEL_3', 'Red Hat Enterprise Linux 3'),
    ('RHEL_3_64', 'Red Hat Enterprise Linux 3 (64 bit)'),
    ('RHEL_4', 'Red Hat Enterprise Linux 4'),
    ('RHEL_4_64', 'Red Hat Enterprise Linux 4 (64 bit)'),
    ('RHEL_5', 'Red Hat Enterprise Linux 5'),
    ('RHEL_5_64', 'Red Hat Enterprise Linux 5 (64 bit) (experimental)'),
    ('RHEL_6', 'Red Hat Enterprise Linux 6'),
    ('RHEL_6_64', 'Red Hat Enterprise Linux 6 (64 bit)'),
    ('RHEL_7', 'Red Hat Enterprise Linux 7'),
    ('RHEL_7_64', 'Red Hat Enterprise Linux 7 (64 bit)'),
    ('CENTOS', 'CentOS 4/5'),
    ('CENTOS_64', 'CentOS 4/5 (64-bit)'),
    ('CENTOS_6', 'CentOS 6'),
    ('CENTOS_6_64', 'CentOS 6 (64-bit)'),
    ('CENTOS_7', 'CentOS 7'),
    ('CENTOS_7_64', 'CentOS 7 (64-bit)'),
    ('ORACLE_LINUX', 'Oracle Linux 4/5'),
    ('ORACLE_LINUX_64', 'Oracle Linux 4/5 (64-bit)'),
    ('ORACLE_LINUX_6', 'Oracle Linux 6'),
    ('ORACLE_LINUX_6_64', 'Oracle Linux 6 (64-bit)'),
    ('ORACLE_LINUX_7', 'Oracle Linux 7'),
    ('ORACLE_LINUX_7_64', 'Oracle Linux 7 (64-bit)'),
    ('SUSE', 'Suse Linux'),
    ('SUSE_64', 'Suse Linux (64 bit)'),
    ('SLES', 'Suse Linux Enterprise Server 9'),
    ('SLES_64', 'Suse Linux Enterprise Server 9 (64 bit)'),
    ('SLES_10', 'Suse linux Enterprise Server 10'),
    ('SLES_10_64', 'Suse Linux Enterprise Server 10 (64 bit) (experimental)'),
    ('SLES_11', 'Suse linux Enterprise Server 11'),
    ('SLES_11_64', 'Suse Linux Enterprise Server 11 (64 bit)'),
    ('SLES_12', 'Suse linux Enterprise Server 12'),
    ('SLES_12_64', 'Suse Linux Enterprise Server 12 (64 bit)'),
    ('NLD_9', 'Novell Linux Desktop 9'),
    ('OES', 'Open Enterprise Server'),
    ('SJDS', 'Sun Java Desktop System'),
    ('MANDRAKE', 'Mandrake Linux'),
    ('MANDRIVA', 'Mandriva Linux'),
    ('MANDRIVA_64', 'Mandriva Linux (64 bit)'),
    ('TURBO_LINUX', 'Turbolinux'),
    ('TURBO_LINUX_64', 'Turbolinux (64 bit)'),
    ('UBUNTU', 'Ubuntu Linux'),
    ('UBUNTU_64', 'Ubuntu Linux (64 bit)'),
    ('DEBIAN_4', 'Debian GNU/Linux 4'),
    ('DEBIAN_4_64', 'Debian GNU/Linux 4 (64 bit)'),
    ('DEBIAN_5', 'Debian GNU/Linux 5'),
    ('DEBIAN_5_64', 'Debian GNU/Linux 5 (64 bit)'),
    ('DEBIAN_6', 'Debian GNU/Linux 6'),
    ('DEBIAN_6_64', 'Debian GNU/Linux 6 (64 bit)'),
    ('DEBIAN_7', 'Debian GNU/Linux 7'),
    ('DEBIAN_7_64', 'Debian GNU/Linux 7 (64 bit)'),
    ('DEBIAN_8', 'Debian GNU/Linux 8'),
    ('DEBIAN_8_64', 'Debian GNU/Linux 8 (64 bit)'),
    ('DEBIAN_9', 'Debian GNU/Linux 9'),
    ('DEBIAN_9_64', 'Debian GNU/Linux 9 (64 bit)'),
    ('DEBIAN_10', 'Debian GNU/Linux 10'),
    ('DEBIAN_10_64', 'Debian GNU/Linux 10 (64 bit)'),
    ('ASIANUX_3', 'Asianux Server 3'),
    ('ASIANUX_3_64', 'Asianux Server 3 (64 bit)'),
    ('ASIANUX_4', 'Asianux Server 4'),
    ('ASIANUX_4_64', 'Asianux Server 4 (64 bit)'),
    ('ASIANUX_5_64', 'Asianux Server 5 (64 bit)'),
    ('ASIANUX_7_64', 'Asianux Server 7 (64 bit)'),
    ('OPENSUSE', 'OpenSUSE Linux'),
    ('OPENSUSE_64', 'OpenSUSE Linux (64 bit)'),
    ('FEDORA', 'Fedora Linux'),
    ('FEDORA_64', 'Fedora Linux (64 bit)'),
    ('COREOS_64', 'CoreOS Linux (64 bit)'),
    ('VMWARE_PHOTON_64', 'VMware Photon (64 bit)'),
    ('OTHER_24X_LINUX', 'Linux 2.4x Kernel'),
    ('OTHER_24X_LINUX_64', 'Linux 2.4x Kernel (64 bit) (experimental)'),
    ('OTHER_26X_LINUX', 'Linux 2.6x Kernel'),
    ('OTHER_26X_LINUX_64', 'Linux 2.6x Kernel (64 bit) (experimental)'),
    ('OTHER_3X_LINUX', 'Linux 3.x Kernel'),
    ('OTHER_3X_LINUX_64', 'Linux 3.x Kernel (64 bit)'),
    ('OTHER_LINUX', 'Linux 2.2x Kernel'),
    ('GENERIC_LINUX', 'Other Linux'),
    ('OTHER_LINUX_64', 'Linux (64 bit) (experimental)'),
    ('SOLARIS_6', 'Solaris 6'),
    ('SOLARIS_7', 'Solaris 7'),
    ('SOLARIS_8', 'Solaris 8'),
    ('SOLARIS_9', 'Solaris 9'),
    ('SOLARIS_10', 'Solaris 10 (32 bit) (experimental)'),
    ('SOLARIS_10_64', 'Solaris 10 (64 bit) (experimental)'),
    ('SOLARIS_11_64', 'Solaris 11 (64 bit)'),
    ('OS2', 'OS/2'),
    ('ECOMSTATION', 'eComStation 1.x'),
    ('ECOMSTATION_2', 'eComStation 2.0'),
    ('NETWARE_4', 'Novell NetWare 4'),
    ('NETWARE_5', 'Novell NetWare 5.1'),
    ('NETWARE_6', 'Novell NetWare 6.x'),
    ('OPENSERVER_5', 'SCO OpenServer 5'),
    ('OPENSERVER_6', 'SCO OpenServer 6'),
    ('UNIXWARE_7', 'SCO UnixWare 7'),
    ('DARWIN', 'Mac OS 10.5'),
    ('DARWIN_64', 'Mac OS 10.5 (64 bit)'),
    ('DARWIN_10', 'Mac OS 10.6'),
    ('DARWIN_10_64', 'Mac OS 10.6 (64 bit)'),
    ('DARWIN_11', 'Mac OS 10.7'),
    ('DARWIN_11_64', 'Mac OS 10.7 (64 bit)'),
    ('DARWIN_12_64', 'Mac OS 10.8 (64 bit)'),
    ('DARWIN_13_64', 'Mac OS 10.9 (64 bit)'),
    ('DARWIN_14_64', 'Mac OS 10.10 (64 bit)'),
    ('DARWIN_15_64', 'Mac OS 10.11 (64 bit)'),
    ('DARWIN_16_64', 'Mac OS 10.12 (64 bit)'),
    ('VMKERNEL', 'VMware ESX 4'),
    ('VMKERNEL_5', 'VMware ESX 5'),
    ('VMKERNEL_6', 'VMware ESX 6'),
    ('VMKERNEL_65', 'VMware ESX 6.5'),
    ('OTHER', 'Other Operating System'),
    ('OTHER_64', 'Other Operating System (64 bit) (experimental)'),
))



from waldur_ansible.playbook_jobs import models, views


def register_in(router):
    router.register(r'ansible-playbooks', views.PlaybookViewSet, base_name=models.Playbook.get_url_name())
    router.register(r'ansible-jobs', views.JobViewSet, base_name=models.Job.get_url_name())



from django.apps import AppConfig
from django.db.models import signals


class MarketplaceAzureConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_azure'
    verbose_name = 'Marketplace Azure'

    def ready(self):
        from waldur_azure import models as azure_models
        from waldur_azure.apps import AzureConfig
        from waldur_mastermind.marketplace.plugins import manager
        from waldur_mastermind.marketplace import handlers as marketplace_handlers

        from . import handlers, processors, VIRTUAL_MACHINE_TYPE, SQL_SERVER_TYPE

        resource_models = (
            azure_models.VirtualMachine,
            azure_models.SQLServer,
            azure_models.SQLDatabase
        )

        marketplace_handlers.connect_resource_handlers(*resource_models)
        marketplace_handlers.connect_resource_metadata_handlers(*resource_models)

        signals.post_save.connect(
            handlers.synchronize_nic,
            sender=azure_models.NetworkInterface,
            dispatch_uid='waldur_mastermind.marketpace_azure.synchronize_nic',
        )

        signals.post_save.connect(
            handlers.synchronize_public_ip,
            sender=azure_models.PublicIP,
            dispatch_uid='waldur_mastermind.marketpace_azure.synchronize_public_ip',
        )

        manager.register(offering_type=VIRTUAL_MACHINE_TYPE,
                         create_resource_processor=processors.VirtualMachineCreateProcessor,
                         delete_resource_processor=processors.VirtualMachineDeleteProcessor,
                         service_type=AzureConfig.service_name)

        manager.register(offering_type=SQL_SERVER_TYPE,
                         create_resource_processor=processors.SQLServerCreateProcessor,
                         delete_resource_processor=processors.SQLServerDeleteProcessor,
                         service_type=AzureConfig.service_name)



from waldur_ansible.python_management import models as python_management_models, utils as python_management_utils
from waldur_ansible.python_management.backend import extracted_information_handlers as python_management_handlers


def set_affected_virtual_env_global_or_not(request, jupyter_hub_global):
    jupyter_hub_management = request.jupyter_hub_management
    affected_virtual_env = python_management_utils.execute_safely(
        lambda: python_management_models.VirtualEnvironment.objects.get(python_management=jupyter_hub_management.python_management, name=request.virtual_env_name))
    if affected_virtual_env:
        affected_virtual_env.jupyter_hub_global = jupyter_hub_global
        affected_virtual_env.save()


def update_installed_libraries_list(lines_post_processor, request):
    installed_libraries_handler = python_management_handlers.InstalledLibrariesExtractedInformationHandler()
    installed_libraries_handler.persist_installed_libraries_in_db(
        request.jupyter_hub_management.python_management, request.virtual_env_name, lines_post_processor.installed_libraries_after_modifications)


class JupyterHubVirtualEnvironmentGlobalExtractedInformationHandler(object):

    def handle_extracted_information(self, request, lines_post_processor):
        set_affected_virtual_env_global_or_not(request, True)
        update_installed_libraries_list(lines_post_processor, request)


class JupyterHubVirtualEnvironmentLocalExtractedInformationHandler(object):
    def handle_extracted_information(self, request, lines_post_processor):
        set_affected_virtual_env_global_or_not(request, False)
        update_installed_libraries_list(lines_post_processor, request)


class JupyterHubManagementDeleteExtractedInformationHandler(object):
    def handle_extracted_information(self, request, lines_post_processor):
        for global_virtual_env in request.jupyter_hub_management.python_management.virtual_environments.filter(jupyter_hub_global=True):
            global_virtual_env.jupyter_hub_global = False
            global_virtual_env.save()
        request.jupyter_hub_management.delete()



from django.apps import AppConfig
from django.db.models import signals


class VMwareConfig(AppConfig):
    name = 'waldur_vmware'
    verbose_name = 'VMware'
    service_name = 'VMware'

    def ready(self):
        from waldur_core.structure import SupportedServices

        from .backend import VMwareBackend
        from . import handlers, models

        SupportedServices.register_backend(VMwareBackend)

        signals.post_save.connect(
            handlers.update_vm_total_disk_when_disk_is_created_or_updated,
            sender=models.Disk,
            dispatch_uid='waldur_vmware.handlers.'
                         'update_vm_total_disk_when_disk_is_created_or_updated',
        )

        signals.post_delete.connect(
            handlers.update_vm_total_disk_when_disk_is_deleted,
            sender=models.Disk,
            dispatch_uid='waldur_vmware.handlers.'
                         'update_vm_total_disk_when_disk_is_deleted',
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.logging.loggers
import django_fsm
import waldur_core.core.models
import django.db.models.deletion
import django.utils.timezone
import waldur_core.core.fields
import taggit.managers
import model_utils.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('openstack', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Volume',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(max_length=150, verbose_name='runtime state', blank=True)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')])),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('size', models.PositiveIntegerField(help_text='Size in MiB')),
                ('bootable', models.BooleanField(default=False)),
                ('metadata', waldur_core.core.fields.JSONField(blank=True)),
                ('image_metadata', waldur_core.core.fields.JSONField(blank=True)),
                ('type', models.CharField(max_length=100, blank=True)),
                ('image', models.ForeignKey(to='openstack.Image', null=True)),
                ('service_project_link', models.ForeignKey(related_name='volumes', on_delete=django.db.models.deletion.PROTECT, to='openstack.OpenStackServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
                ('tenant', models.ForeignKey(related_name='volumes', to='openstack.Tenant')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
    ]
